/*
 * audio.h: Audio frame handling
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#ifndef _AUDIO_H_
#define _AUDIO_H_

#include "stream.h"
#include <stdio.h>

extern "C" {
#include <ffmpeg/avcodec.h>
}

extern FILE *rawFile;
extern FILE *mpegFile;
extern FILE *ac3File;

/**
 * A class representing an uncompressed audio frame.
 */
class cRawAudioFrame
{
public:
	/**
	 * Constructs the audio frame.
	 *
	 * @param data pointer to frame data
	 * @param length length of frame data
	 * @param channels number of channels of the compressed audio frame
	 */
	cRawAudioFrame(unsigned char * data, unsigned int length, int channels);
	/**
	 * Destructs the audio frame.
	 */
	~cRawAudioFrame();
	/**
	 * Retrieves the data of the audio frame.
	 *
	 * @return a pointer to the data of the audio frame
	 */
	unsigned char * Data() { return data; }
	/**
	 * Retrieves the length of the audio frame.
	 *
	 * @return length in bytes
	 */
	unsigned int Length() { return length; }
	/**
	 * Retrieves the number of channels of the compressed audio frame.
	 *
	 * @return number of channels of the compressed audio frame
	 */
	int Channels() { return channels; }
private:
	unsigned char *data;
	unsigned int length;
	int channels;
};

/**
 * A class representing a compressed MPEG audio frame.
 */
class cMPEGAudioFrame
:	public cBitStream
{
public:    
	/**
	 * Constructs the audio frame. Parses the data and fills header structures.
	 *
	 * @param data pointer to frame data
	 * @param length length of frame data
	 */
	cMPEGAudioFrame(unsigned char * data, unsigned int length);
	/**
	 * Destructs the audio frame. Actually does nothing.
	 */
	~cMPEGAudioFrame();
	/**
	 * Decodes the audio frame using libavcodec.
	 *
	 * @return the decompressed audio frame
	 */
	cRawAudioFrame * Decode(AVCodecContext * c);
	/**
	 * Retrieves the size of the audio frame.
	 *
	 * @return size of the audio frame in bytes
	 */
	int FrameSize() { return frameSize; }
	/**
	 * Retrieves the MPEG version of the audio frame.
	 *
	 * @return MPEG version of the audio frame (0 = MPEG 2.5, 2 = MPEG 2, 3 = MPEG 1)
	 */
	int Version() { return version; }
	/**
	 * Retrieves the MPEG Audio Layer of the audio frame.
	 *
	 * @return MPEG Audio Layer of the audio frame
	 */
	int Layer() { return layer; }
	/**
	 * Retrieves the bitrate of the audio frame.
	 *
	 * @return bitrate of the audio frame in kbits per second
	 */
	int BitRate() { return bitRate; }
	/**
	 * Retrieves the sampling rate of the audio frame.
	 *
	 * @return sampling rate of the audio frame
	 */
	int SamplingRate() { return samplingRate; }
	/**
	 * Retrieves the padding bit of the audio frame.
	 *
	 * @return padding bit of the audio frame
	 */
	int Padding() { return padding; }
	/**
	 * Retrieves the channel mode of the audio frame.
	 *
	 * @return channel mode of the audio frame (0 = Stereo, 1 = Joint Stereo, 2 = Dual Channel, 3 = Single Channel(Mono))
	 */
	int ChannelMode() { return channelMode; }
	/**
	 * Retrieves the channel mode extension of the audio frame.
	 *
	 * @return channel mode extension of the audio frame
	 */
	int ModeExtension() { return modeExtension; }
	/**
	 * Retrieves the copyright bit of the audio frame.
	 *
	 * @return copyright bit of the audio frame
	 */
	int Copyright() { return copyright; }
	/**
	 * Retrieves the original bit of the audio frame.
	 *
	 * @return original bit of the audio frame
	 */
	int Original() { return original; }
	/**
	 * Retrieves the emphasis of the audio frame.
	 *
	 * @return emphasis of the audio frame (0 = none, 1 = 50/15ms, 3 = CCIT J.17)
	 */
	int Emphasis() { return emphasis; }
private:
	void Parse();
	int frameSize;
	int version;
	int layer;
	int bitRate;
	int samplingRate;
	int padding;
	int channelMode;
	int modeExtension;
	int copyright;
	int original;
	int emphasis;
};

/**
 * A class representing a compressed AC3 audio frame.
 */
class cAC3AudioFrame
:	public cBitStream
{
public:    
	/**
	 * Constructs the audio frame. Parses the data and fills header structures.
	 *
	 * @param data pointer to frame data
	 * @param length length of frame data
	 */
    cAC3AudioFrame(unsigned char * data, unsigned int length);
	/**
	 * Destructs the audio frame. Actually does nothing.
	 */
	~cAC3AudioFrame();
	/**
	 * Retrieves the size of the audio frame.
	 *
	 * @return size of the audio frame in bytes
	 */
	int FrameSize() { return frameSize; }
	/**
	 * Retrieves the bitrate of the audio frame.
	 *
	 * @return bitrate of the audio frame in kbits per second
	 */
	int BitRate() { return bitRate; }
	/**
	 * Retrieves sampling rate bit of the audio frame.
	 *
	 * @return sampling rate of the audio frame
	 */
	int SamplingRate() { return samplingRate; }
	/**
	 * Retrieves the channel mode of the audio frame. See constants kA52_... in audio.h for possible modes.
	 *
	 * @return channel mode of the audio frame
	 */
	int ChannelMode() { return channelMode; }
private:
	void Parse();
	int frameSize;
	int bitRate;
	int samplingRate;
	int channelMode;
};

const unsigned short kBitrateTab[2][3][15] =
{
	{
		{0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448 },
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384 },
		{0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320 }
	},
	{
		{0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256},
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160},
		{0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160}
	}
};

const unsigned short kFreqTab[3] =
{
	44100,
	48000,
	32000
};

const int kA52_Channel		= 0;
const int kA52_Mono			= 1;
const int kA52_Stereo		= 2;
const int kA52_3F			= 3;
const int kA52_2F1R			= 4;
const int kA52_3F1R			= 5;
const int kA52_2F2R			= 6;
const int kA52_3F2R			= 7;
const int kA52_Channel1		= 8;
const int kA52_Channel2		= 9;
const int kA52_Dolby		= 10;
const int kA52_ChannelMask	= 15;
const int kA52_LFE			= 16;
const int kA52_AdjustLevel	= 32;

#endif

