/*
 * feature.cpp: Features for commercial detection
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#include "global.h"
#include "feature.h"
#include "util.h"

extern "C" {
#include <ffmpeg/i386/mmx.h>
}

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <errno.h>

char cFeature::dump[256] = "";

const int cBorderFeature::kThresholdBinarization = 128;
const int cBorderFeature::kMaxHorz = 100;
const int cBorderFeature::kMaxVert = 50;
const int cBorderFeature::kMaxLineLuminance = 35;
const int cBorderFeature::kMinPictureLuminance = 35;
const int cBorderFeature::kFactor = 20;

cBorderFeature::cBorderFeature()
:	firstLine(0),
	lastLine(0),
	firstCol(0),
	lastCol(0),
	luminance(0)
{
	this->type = kFeatureTypeRawVideo;
	this->oldBorder.top = 0;
	this->oldBorder.bottom = 0;
	this->oldBorder.left = 0;
	this->oldBorder.right = 0;
	this->newBorder.top = 0;
	this->newBorder.bottom = 0;
	this->newBorder.left = 0;
	this->newBorder.right = 0;
}

cBorderFeature::~cBorderFeature()
{
}

void cBorderFeature::Refresh(void * frame)
{
	int onePixelChange = 0;
	int twoPixelChange = 0;
	int threePixelChange = 0;
	int morePixelChange = 0;
	int pixelChange[4];
	int i;
	
	this->frame = (cRawVideoFrame *) frame;
	this->luminance = CheckLuminance();
	if (this->luminance > kMinPictureLuminance)
	{
		this->CheckHorizontal(true, 0);
		this->CheckHorizontal(false, this->frame->Height() - 1);
		this->CheckVertical(true, 0);
		this->CheckVertical(false, this->frame->Width() - 1);
	}
	pixelChange[0] = abs(this->oldBorder.top - this->newBorder.top);
	pixelChange[1] = abs(this->oldBorder.bottom - this->newBorder.bottom);
	pixelChange[2] = abs(this->oldBorder.left - this->newBorder.left);
	pixelChange[3] = abs(this->oldBorder.right - this->newBorder.right);
	for (i = 0; i < 4; i++)
	{
		if (pixelChange[i] > 0)
		{
			if (pixelChange[i] > 1)
			{
				if (pixelChange[i] > 2)
				{
					if (pixelChange[i] > 3)
					{
						morePixelChange++;
					}
					else
					{
						threePixelChange++;
					}
				}
				else
				{
					twoPixelChange++;
				}
			}
			else
			{
				onePixelChange++;
			}
		}
	}
	if (morePixelChange > 0 ||
//		threePixelChange > 1 ||
		twoPixelChange > 2)
	{
		this->changed = true;
		this->oldBorder.top = this->newBorder.top;
		this->oldBorder.bottom = this->newBorder.bottom;
		this->oldBorder.left = this->newBorder.left;
		this->oldBorder.right = this->newBorder.right;
	}
	else
	{
		this->changed = false;
	}
}

char * cBorderFeature::Dump()
{
	snprintf(dump, 255, "%4d%4d%4d%4d    %4d%4d%4d%4d    %4d",
			 this->oldBorder.top, this->oldBorder.bottom,
			 this->oldBorder.left, this->oldBorder.right,
			 this->firstLine, this->lastLine,
			 this->firstCol, this->lastCol,
			 this->luminance);
	return dump;
}

void cBorderFeature::CheckHorizontal(bool top, int start)
{
	int width = this->frame->Width();
	int lineSize = this->frame->LineSize(0);
	unsigned char *data = this->frame->PictureData(0);
	int column;
	int line;
	int lumi;
	int oldLumi = 0;
	int startPos;
	int end = top ? kMaxHorz : start - kMaxHorz;
	int increment = top ? 1 : -1;
	int black;
	double ratio;
	double oldRatio = 0.0;

	line = start;
	while (line != end)
	{
		lumi = 0;
		black = 0;
		startPos = line * lineSize;
		for (column = 0; column < width; column++)
		{
			lumi += data[startPos + column];
			if (data[startPos + column] < 21)
				black++;
		}
		lumi /= width;
		ratio = (double) black / (double) width;
		if (line == start)
		{
			if (top)
				firstLine = lumi;
			else
				lastLine = lumi;
			if (lumi > kMaxLineLuminance)
				break;
		}
		else
		{
			if (oldRatio / ratio > 2.0)
				break;
			if (lumi > (kFactor * oldLumi) / 10)
				break;
			if (ratio < 0.7)
			{
				line = end;
				break;
			}
		}
		oldRatio = ratio;
		oldLumi = lumi;
		line += increment;
	}
	if (line != end)
	{
		if (top)
			this->newBorder.top = line;
		else
			this->newBorder.bottom = start - line;
	}
}

void cBorderFeature::CheckVertical(bool left, int start)
{
	int height = this->frame->Height();
	int lineSize = this->frame->LineSize(0);
	unsigned char *data = this->frame->PictureData(0);
	int column;
	int line;
	int lumi;
	int oldLumi = 0;
	int startPos;
	int end = left ? kMaxVert : start - kMaxVert;
	int increment = left ? 1 : -1;
	int black;
	double ratio;
	double oldRatio = 0.0;

	column = start;
	while (column != end)
	{
		lumi = 0;
		black = 0;
		for (line = 0; line < height; line++)
		{
			startPos = line * lineSize;
			lumi += data[startPos + column];
			if (data[startPos + column] < 21)
				black++;
		}
		lumi /= height;
		ratio = ((double) black) / ((double) height);
		if (column == start)
		{
			if (left)
				firstCol = lumi;
			else
				lastCol = lumi;
			if (lumi > kMaxLineLuminance)
				break;
		}
		else
		{
			if (oldRatio / ratio > 2.0)
				break;
			if (lumi > (kFactor * oldLumi) / 10)
				break;
			if (ratio < 0.7)
			{
				column = end;
				break;
			}
		}
		oldRatio = ratio;
		oldLumi = lumi;
		column += increment;
	}
	if (column != end)
	{
		if (left)
			this->newBorder.left = column;
		else
			this->newBorder.right = start - column;
	}
}

int cBorderFeature::CheckLuminance()
{
	unsigned char * picture = this->frame->PictureData(0);
	int lineSize = this->frame->LineSize(0);
	int width = this->frame->Width();
	int height = this->frame->Height();
	int row, column;
	int rowmin, rowmax, colmin, colmax;
	int lumi = 0;
	int pixels = 0;

	rowmin = 150;
	rowmax = height - 150;
	colmin = 20;
	colmax = width - 20;
	for (row = rowmin; row < rowmax; row++)
	{
		for (column = colmin; column < colmax; column++)
		{
			lumi += picture[row * lineSize + column];
			pixels++;
		}
	}
	lumi = lumi / pixels;
	return lumi;
}

cLogoTemplate::cLogoTemplate(char * name, int station, char * edgefile, char * maskfile)
:	station(station),
	width(0),
	height(0),
	normalFactor(0)
{
	strncpy(this->name, name, 100);
	memset(this->data, 0, 10000);
	memset(this->mask, 0, 10000);
	LoadPGM(this->data, edgefile);
	LoadPGM(this->mask, maskfile);

	int i;
	int size = this->width * this->height;
	for (i = 0; i < size; i++)
	{
		if (this->data[i] == 255)
		{
			this->data[i] = 1;
			this->normalFactor++;
		}
		else
		{
			this->data[i] = 0;
		}
		if (this->mask[i] == 255)
		{
			this->mask[i] = 1;
		}
		else
		{
			this->mask[i] = 0;
		}
	}
	if (this->normalFactor == 0)
		this->normalFactor = 1;
}

cLogoTemplate::~cLogoTemplate()
{
}

void cLogoTemplate::AddPosition(int x, int y)
{
	tPosition pos;
	pos.x = x;
	pos.y = y;
	this->positions.push_back(pos);
}

void cLogoTemplate::LoadPGM(unsigned char * buffer, char * filename)
{
	FILE *fp;
	unsigned char tmpbuffer[10000];
	char numbers[100];
	int length;
	int start;
	int pos;
	int line;
	char file[1000];

	sprintf(file, "%s/logos/%s", kFilePath, filename);
	fp = fopen(file, "rb");
	if (fp)
	{
		length = fread(tmpbuffer, 1, 10000, fp);
		pos = 0;
		line = 0;
		while (pos < length)
		{
			if (tmpbuffer[pos] == '#')
			{
				pos = NextLine(tmpbuffer, pos);
			}
			else
			{
				if (line == 0)
				{
					// Magic Number
					pos = NextLine(tmpbuffer, pos);
					line++;
					continue;
				}
				if (line == 1)
				{
					// Size
					start = pos;
					pos = NextLine(tmpbuffer, pos);
					if (pos > length)
						break;
					memcpy(numbers, &tmpbuffer[start], pos - start);
					numbers[pos - start] = 0;
					if (sscanf(numbers, "%d %d", &this->width, &this->height) != 2)
						break;
					line++;
					continue;
				}
				if (line == 2)
				{
					// max Gray
					pos = NextLine(tmpbuffer, pos);
					line++;
					continue;
				}
				if (line == 3)
				{
					// Data
					if (pos < length)
					{
						memcpy(buffer, &tmpbuffer[pos], length - pos);
					}
					break;
				}
			}
		}
		fclose(fp);
	}
}

int cLogoTemplate::NextLine(unsigned char * buffer, int pos)
{
	while (buffer[pos] != 0x0A && pos < 10000)
		pos++;
	pos++;
	return pos;
}

const int cLogoFeature::kThresholdBinarization = 127;
const int cLogoFeature::kThresholdLogoMatching = 200;
const int cLogoFeature::kMinLogoMaskLuminance = 125;
const int cLogoFeature::kMaxEdgeMaskLuminance = 160;
const int cLogoFeature::kMaxNoLogoMaskLuminance = 100;
const int cLogoFeature::kBorder = 4;

cLogoFeature::cLogoFeature()
:	oldLogoIndex(-2),
	newLogoIndex(-2),
	station(-1)
{
	this->type = kFeatureTypeRawVideo;
	memset(this->edgePicture, 0, 512 * 1024);
	this->logos.push_back(new cLogoTemplate("Kabel1", 0, "kabel1.pgm", "kabel1m.pgm"));
	this->logos[0]->AddPosition(55, 65);
	this->logos.push_back(new cLogoTemplate("Kabel1k", 0, "kabel1k.pgm", "kabel1km.pgm"));
	this->logos[1]->AddPosition(52, 115);
	this->logos.push_back(new cLogoTemplate("Pro7", 1, "pro7.pgm", "pro7m.pgm"));
	this->logos[2]->AddPosition(650, 58);
	this->logos[2]->AddPosition(650, 108);
	this->logos.push_back(new cLogoTemplate("RTL", 2, "rtl.pgm", "rtlm.pgm"));
	this->logos[3]->AddPosition(72, 52);
}

cLogoFeature::~cLogoFeature()
{
	int i;
	for (i = 0; i < (int) this->logos.size(); i++)
	{
		cLogoTemplate * f = this->logos[i];
		delete f;
	}
}

void cLogoFeature::Refresh(void * frame)
{
	this->frame = (cRawVideoFrame *) frame;
	CheckLogo();
	emms();
	if (this->oldLogoIndex != this->newLogoIndex ||
		abs(this->oldPos.x - this->newPos.x) > 1 ||
		abs(this->oldPos.y - this->newPos.y) > 1)
	{
		this->changed = true;
		this->oldPos.x = this->newPos.x;
		this->oldPos.y = this->newPos.y;
		this->oldLogoIndex = this->newLogoIndex;
	}
	else
	{
		this->changed = false;
	}
}

char * cLogoFeature::Dump()
{
	snprintf(dump, 255, "%s %4d%4d %4d %4d",
			 this->newLogoIndex >= 0 ? this->logos[this->newLogoIndex]->Name() : "no Logo",
			 this->newPos.x, this->newPos.y,
			 this->edgeMaskLuminance, this->logoMaskLuminance);
	return dump;
}

void cLogoFeature::CheckLogo()
{
	unsigned char * picture = this->frame->PictureData(0);
	int lineSize = this->frame->LineSize(0);
	int width = this->frame->Width();
	int height = this->frame->Height();

	CalcEdges(picture, lineSize, width, height);

	cLogoTemplate *logo;
	int logoIndex;
	int maxval;
	int lumi;
	tPosition logoPos;

	if (this->newLogoIndex >= 0)
	{
		logo = this->logos[this->newLogoIndex];
		maxval = 0;
		logoPos = this->newPos;
		if (MatchLogoPosition(lineSize, width, height, logo, &logoPos, &maxval))
		{
			lumi = MaskLuminance(this->edgePicture, lineSize, width, height, logo, logoPos);
			this->edgeMaskLuminance = lumi;
			if (lumi < kMaxEdgeMaskLuminance)
			{
				lumi = MaskLuminance(picture, lineSize, width, height, logo, logoPos);
				this->logoMaskLuminance = lumi;
				if (lumi > kMaxNoLogoMaskLuminance)
				{
					this->newPos.x = logoPos.x;
					this->newPos.y = logoPos.y;
				}
			}
		}
		else
		{
			lumi = MaskLuminance(picture, lineSize, width, height, logo, this->newPos);
			this->logoMaskLuminance = lumi;
			if (lumi < kMinLogoMaskLuminance)
				this->newLogoIndex = -1;
		}
	}
	else
	{
		for (logoIndex = 0; logoIndex < (int) this->logos.size(); logoIndex++)
		{
			logo = this->logos[logoIndex];
			if (this->station >= 0)
				if (logo->Station() != this->station)
					continue;
			maxval = 0;
			logoPos.x = 0;
			logoPos.y = 0;

			if (MatchLogo(lineSize, width, height, logo, &logoPos, &maxval))
			{
				lumi = MaskLuminance(edgePicture, lineSize, width, height, logo, logoPos);
				this->edgeMaskLuminance = lumi;
				if (lumi < kMaxEdgeMaskLuminance)
				{
					lumi = MaskLuminance(picture, lineSize, width, height, logo, logoPos);
					this->logoMaskLuminance = lumi;
					if (lumi > kMaxNoLogoMaskLuminance)
					{
						this->newPos.x = logoPos.x;
						this->newPos.y = logoPos.y;
						this->newLogoIndex = logoIndex;
						this->station = logo->Station();
					}
				}
			}
			else
			{
				this->newLogoIndex = -1;
			}
		}
	}
/*	FILE *fp;
	fp = fopen("vidtempl.raw", "wb");
	for (int i = 0; i < height; i++)
	{
		fwrite(&outbuf[i * lineSize], width, 1, fp);
	}
	fclose(fp);
/**/
}

void cLogoFeature::CalcEdges(unsigned char * picture, int lineSize, int width, int height)
{
	int row, column;
	int rowmin, rowmax, colmin, colmax;
	int sum;
	int sum1;
	int sum2;
	int pos;
	register unsigned char * prevline;
	register unsigned char * currline;
	register unsigned char * nextline;

	rowmin = 1;
	rowmax = rowmin + 150;
	colmin = 1;
	colmax = width - 1;
	for (row = rowmin; row < rowmax; row++)
	{
		pos = row * lineSize + colmin;
		currline = &picture[row * lineSize];
		prevline = currline - lineSize;
		nextline = currline + lineSize;
		for (column = colmin; column < colmax; column++)
		{
			// Sobel
			sum1 = 0;
			sum1 -= *prevline;
			sum1 += *(nextline + 2);
			sum2 = sum1;
			sum1 -= *(prevline + 1) << 1;
			sum1 -= *(prevline + 2);
			sum1 += *nextline;
			sum1 += *(nextline + 1) << 1;
			if (sum1 < 0)
				sum1 = -sum1;

			sum2 += *(prevline + 2);
			sum2 -= *currline << 1;
			sum2 += *(currline + 2) << 1;
			sum2 -= *nextline;
			if (sum2 < 0)
				sum2 = -sum2;
			sum = sum1 + sum2;

			if (sum > kThresholdBinarization)
				sum = 255;
			else
				sum = 0;
			this->edgePicture[pos] = sum;
			pos++;
			prevline++;
			currline++;
			nextline++;
		}
	}
	rowmax = height - 1;
	rowmin = rowmax - 150;
	colmin = 1;
	colmax = width - 1;
	for (row = rowmin; row < rowmax; row++)
	{
		pos = row * lineSize + colmin;
		currline = &picture[row * lineSize];
		prevline = currline - lineSize;
		nextline = currline + lineSize;
		for (column = colmin; column < colmax; column++)
		{
			// Sobel
			sum1 = 0;
			sum1 -= *prevline;
			sum1 += *(nextline + 2);
			sum2 = sum1;
			sum1 -= *(prevline + 1) << 1;
			sum1 -= *(prevline + 2);
			sum1 += *nextline;
			sum1 += *(nextline + 1) << 1;
			if (sum1 < 0)
				sum1 = -sum1;

			sum2 += *(prevline + 2);
			sum2 -= *currline << 1;
			sum2 += *(currline + 2) << 1;
			sum2 -= *nextline;
			if (sum2 < 0)
				sum2 = -sum2;
			sum = sum1 + sum2;

			if (sum > kThresholdBinarization)
				sum = 255;
			else
				sum = 0;
			this->edgePicture[pos] = sum;
			pos++;
			prevline++;
			currline++;
			nextline++;
		}
	}
/*	FILE *fp;
	fp = fopen("videdge.raw", "wb");
	for (int i = 0; i < height; i++)
	{
		fwrite(&edgePicture[i * lineSize], width, 1, fp);
	}
	fclose(fp);
/**/
}

bool cLogoFeature::MatchLogo(int lineSize, int width, int height,
							 cLogoTemplate *logo,
							 tPosition *retPos, int *retMaxVal)
{
	unsigned char *logoData;
	int position, row, column, i, j;
	int kernel_rows, kernel_cols, dead_rows, dead_cols;
	int rowmin, rowmax, colmin, colmax, jmax;
	int normal_factor;
	int sumval;
	int pixval;
	int bufpos;
	int outpos;
	register unsigned int * ldataptr;
	register unsigned int * bdataptr;
	unsigned int temp;
	int maxval;
	tPosition logoPos;
	tPosition logoPosFound;

	logoData = logo->Data();
	kernel_rows = logo->Height();
	kernel_cols = logo->Width();
	dead_rows = kernel_rows / 2;
	dead_cols = kernel_cols / 2;
	normal_factor = logo->NormalFactor();
	maxval = 0;
	logoPosFound.x = 0;
	logoPosFound.y = 0;

	for (position = 0; position < logo->PositionCount(); position++)
	{
		logoPos = logo->Position(position);
		rowmin = logoPos.y - dead_rows - kBorder;
		if (rowmin < 0)
			rowmin = 0;
		rowmax = rowmin + (2 * kBorder);
		if (rowmax > (height - kernel_rows + 1))
			rowmax = (height - kernel_rows + 1);
		colmin = logoPos.x - dead_cols - kBorder;
		if (colmin < 0)
			colmin = 0;
		colmax = colmin + (2 * kBorder);
		if (colmax > (width - kernel_cols + 1))
			colmax = (width - kernel_cols + 1);
		jmax = kernel_cols / 4;

		for (row = rowmin; row < rowmax; row++)
		{
			outpos = (row + dead_rows) * lineSize + dead_cols + colmin;
			for (column = colmin; column < colmax; column++)
			{
				sumval = 0;
				ldataptr = (unsigned int *) &logoData[0];
				bufpos = row * lineSize + column;
				for (i = 0; i < kernel_rows; i++)
				{
					bdataptr = (unsigned int *) &this->edgePicture[bufpos];
					for (j = 0; j < jmax; j++)
					{
						// MMX
						pxor_r2r(mm2,mm2);
						movd_m2r(*ldataptr,mm0);
						movd_m2r(*bdataptr,mm1);
						punpcklbw_r2r(mm2,mm0);
						punpcklbw_r2r(mm2,mm1);
						pmaddwd_r2r(mm1,mm0);
						movq_r2r(mm0,mm1);
						psrlq_i2r(32,mm1);
						paddd_r2r(mm0,mm1);
						movd_r2m(mm1,temp);
						sumval += temp;
						bdataptr++;
						ldataptr++;
					}
					bufpos += lineSize;
				}
				if ((pixval = abs (sumval / normal_factor)) > 255)
					pixval = 255;
				outbuf[outpos] = (unsigned char) pixval;
				outpos++;
				if (pixval > maxval)
				{
					logoPosFound.x = column;
					logoPosFound.y = row;
					maxval = pixval;
				}
			}
		}
	}
	logoPosFound.x += dead_cols;
	logoPosFound.y += dead_rows;
	*retMaxVal = maxval;
	*retPos = logoPosFound;
	if (maxval > kThresholdLogoMatching)
	{
		return true;
	}
	return false;
}

bool cLogoFeature::MatchLogoPosition(int lineSize, int width, int height,
									 cLogoTemplate *logo,
									 tPosition *retPos, int *retMaxVal)
{
	unsigned char *logoData;
	int row, column, i, j;
	int kernel_rows, kernel_cols, dead_rows, dead_cols;
	int rowmin, rowmax, colmin, colmax, jmax;
	int normal_factor;
	int sumval;
	int pixval;
	int bufpos;
	int outpos;
	register unsigned int * ldataptr;
	register unsigned int * bdataptr;
	unsigned int temp;
	int maxval;
	tPosition logoPos;

	logoData = logo->Data();
	kernel_rows = logo->Height();
	kernel_cols = logo->Width();

	dead_rows = kernel_rows / 2;
	dead_cols = kernel_cols / 2;

	normal_factor = logo->NormalFactor();
	logoPos = *retPos;
	rowmin = logoPos.y - dead_rows - kBorder;
	if (rowmin < 0)
		rowmin = 0;
	rowmax = rowmin + (2 * kBorder);
	if (rowmax > (height - kernel_rows + 1))
		rowmax = (height - kernel_rows + 1);
	colmin = logoPos.x - dead_cols - kBorder;
	if (colmin < 0)
		colmin = 0;
	colmax = colmin + (2 * kBorder);
	if (colmax > (width - kernel_cols + 1))
		colmax = (width - kernel_cols + 1);
	jmax = kernel_cols / 4;

	maxval = 0;
	logoPos.x = 0;
	logoPos.y = 0;

	for (row = rowmin; row < rowmax; row++)
	{
		outpos = (row + dead_rows) * lineSize + dead_cols + colmin;
		for (column = colmin; column < colmax; column++)
		{
			sumval = 0;
			ldataptr = (unsigned int *) &logoData[0];
			bufpos = row * lineSize + column;
			for (i = 0; i < kernel_rows; i++)
			{
				bdataptr = (unsigned int *) &this->edgePicture[bufpos];
				for (j = 0; j < jmax; j++)
				{
					// MMX
					pxor_r2r(mm2,mm2);
					movd_m2r(*ldataptr,mm0);
					movd_m2r(*bdataptr,mm1);
					punpcklbw_r2r(mm2,mm0);
					punpcklbw_r2r(mm2,mm1);
					pmaddwd_r2r(mm1,mm0);
					movq_r2r(mm0,mm1);
					psrlq_i2r(32,mm1);
					paddd_r2r(mm0,mm1);
					movd_r2m(mm1,temp);
					sumval += temp;
					bdataptr++;
					ldataptr++;
				}
				bufpos += lineSize;
			}
			if ((pixval = abs (sumval / normal_factor)) > 255)
				pixval = 255;
			outbuf[outpos] = (unsigned char) pixval;
			outpos++;
			if (pixval > maxval)
			{
				logoPos.x = column;
				logoPos.y = row;
				maxval = pixval;
			}
		}
	}
	logoPos.x += dead_cols;
	logoPos.y += dead_rows;

	*retMaxVal = maxval;
	*retPos = logoPos;
	if (maxval > kThresholdLogoMatching)
	{
		return true;
	}
	return false;
}

int cLogoFeature::MaskLuminance(unsigned char * picture, int lineSize,
								int width, int height, cLogoTemplate *logo,
								tPosition logoPos)
{
	unsigned char *logoMask;
	int i, j;
	int kernel_rows, kernel_cols;
	int rowmin, colmin;
	int sumval;
	int pixels;
	int bufpos;
	register unsigned char * lmaskptr;
	register unsigned char * bdataptr;

	logoMask = logo->Mask();
	kernel_rows = logo->Height();
	kernel_cols = logo->Width();

	rowmin = logoPos.y - kernel_rows / 2;
	if (rowmin < 0)
		rowmin = 0;
	colmin = logoPos.x - kernel_cols / 2;
	if (colmin < 0)
		colmin = 0;
	sumval = 0;
	pixels = 0;
	lmaskptr = (unsigned char *) &logoMask[0];
	bufpos = rowmin * lineSize + colmin;
	for (i = 0; i < kernel_rows; i++)
	{
		bdataptr = (unsigned char *) &picture[bufpos];
		for (j = 0; j < kernel_cols; j++)
		{
			if (*lmaskptr)
			{
				sumval += *bdataptr;
				pixels++;
			}
			lmaskptr++;
			bdataptr++;
		}
		bufpos += lineSize;
	}
	if (pixels == 0)
		pixels = 1;
	return sumval / pixels;
}


const int cBlackWhiteFeature::kBlackWhite = 0;
const int cBlackWhiteFeature::kColour = 1;
const int cBlackWhiteFeature::kMinChrominance = 125;
const int cBlackWhiteFeature::kMaxChrominance = 131;

cBlackWhiteFeature::cBlackWhiteFeature()
:	oldState(-1),
	newState(-1)
{
	this->type = kFeatureTypeRawVideo;
}

cBlackWhiteFeature::~cBlackWhiteFeature()
{
}

void cBlackWhiteFeature::Refresh(void * frame)
{
	this->frame = (cRawVideoFrame *) frame;
	Check();
	if (this->oldState != this->newState)
	{
		this->changed = true;
		this->oldState = this->newState;
	}
	else
		this->changed = false;
}

char * cBlackWhiteFeature::Dump()
{
	snprintf(dump, 255, "%d", this->newState);
	return dump;
}

void cBlackWhiteFeature::Check()
{
	unsigned char * picture;
	int lineSize;
	int width = this->frame->Width() / 2;
	int height = this->frame->Height() / 2;
	int row, column;
	int rowmin, rowmax, colmin, colmax;
	int pixels, pixelsBW;
	int value;
	double ratio;

	rowmin = 70;
	rowmax = height - 1 - rowmin;
	colmin = 10;
	colmax = width - 1 - colmin;

	picture = this->frame->PictureData(1);
	lineSize = this->frame->LineSize(1);
	pixels = 0;
	pixelsBW = 0;
	for (row = rowmin; row < rowmax; row++)
	{
		for (column = colmin; column < colmax; column++)
		{
			value = picture[row * lineSize + column];
			if (value > kMinChrominance && value < kMaxChrominance)
				pixelsBW++;
			pixels++;
		}
	}
	ratio = (double) pixelsBW / (double) pixels;
	if (ratio < 0.90)
	{
		this->newState = kColour;
		return;
	}

	picture = this->frame->PictureData(2);
	lineSize = this->frame->LineSize(2);
	pixels = 0;
	pixelsBW = 0;
	for (row = rowmin; row < rowmax; row++)
	{
		for (column = colmin; column < colmax; column++)
		{
			value = picture[row * lineSize + column];
			if (value > kMinChrominance && value < kMaxChrominance)
				pixelsBW++;
			pixels++;
		}
	}
	ratio = (double) pixelsBW / (double) pixels;
	if (ratio < 0.90)
	{
		this->newState = kColour;
		return;
	}
	this->newState = kBlackWhite;
}

cAC3ChannelFeature::cAC3ChannelFeature()
:	oldChannelMode(0),
	newChannelMode(0)
{
	this->type = kFeatureTypeAC3Audio;
}

cAC3ChannelFeature::~cAC3ChannelFeature()
{
}

void cAC3ChannelFeature::Refresh(void * frame)
{
	this->frame = (cAC3AudioFrame *) frame;
	CheckChannels();
	if (this->oldChannelMode != this->newChannelMode)
	{
		this->changed = true;
		this->oldChannelMode = this->newChannelMode;
	}
	else
		this->changed = false;
}

char * cAC3ChannelFeature::Dump()
{
	snprintf(dump, 255, "%d", this->newChannelMode);
	return dump;
}

void cAC3ChannelFeature::CheckChannels()
{
	this->newChannelMode = frame->ChannelMode();
}

cMonoStereoFeature::cMonoStereoFeature()
:	oldChannels(0),
	newChannels(0)
{
	this->type = kFeatureTypeRawAudio;
}

cMonoStereoFeature::~cMonoStereoFeature()
{
}

void cMonoStereoFeature::Refresh(void * frame)
{
	this->frame = (cRawAudioFrame *) frame;
	CheckChannels();
	if (this->oldChannels != this->newChannels)
	{
		this->changed = true;
	}
	else
		this->changed = false;
	this->oldChannels = this->newChannels;
}

char * cMonoStereoFeature::Dump()
{
	snprintf(dump, 255, "%3d", this->oldChannels);
	return dump;
}

void cMonoStereoFeature::CheckChannels()
{
	unsigned char *data;
	unsigned int length;
	unsigned int pos;
	int chan1;
	int chan2;
	int mono;
	int samples;
	double ratio;
	this->newChannels = frame->Channels();
	if (this->newChannels == 2)
	{
		data = this->frame->Data();
		length = this->frame->Length();
		pos = 0;
		mono = 0;
		samples = 0;
		while (pos < length)
		{
			chan1 = *(short *) &data[pos];
			pos += 2;
			chan2 = *(short *) &data[pos];
			pos += 2;
			if (abs(chan1 - chan2) < 200)
				mono++;
			samples++;
		}
		ratio = (double) mono / (double) samples;
		if (ratio > 0.8)
			this->newChannels = 1;
		else
			this->newChannels = 2;
	}
}
