/*
 * feature.h: Features for commercial detection
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#ifndef _FEATURE_H_
#define _FEATURE_H_

#include <vector>
using namespace std;

#include "audio.h"
#include "video.h"

const int kFeatureTypeNone		= 0;
const int kFeatureTypePES		= 1;
const int kFeatureTypeMPEGAudio	= 2;
const int kFeatureTypeMPEGVideo	= 3;
const int kFeatureTypeAC3Audio	= 4;
const int kFeatureTypeRawAudio	= 5;
const int kFeatureTypeRawVideo	= 6;

/**
 * Base class for all features. A feature is a characteristic of the DVB-stream
 * which helps to differentiate between commercial and ordinary programme. This
 * can be characteristics in audio or video signals, as well as in the headers
 * of the various streams.
 */
class cFeature
{
public:
	/**
	 * Constructs the feature.
	 */
	cFeature() { type = kFeatureTypeNone; changed = false; }
	/**
	 * Destructs the feature. Actually does nothing.
	 */
	virtual ~cFeature() {}
	/**
	 * Refreshs the feature.
	 *
	 * @param frame the frame, which data the feature is based on
	 */
	virtual void Refresh(void * frame) {}
	/**
	 * Dumps the feature data to a string. Usefull for logging. This method has to be reimplemented by subclasses.
	 *
	 * @return null-terminated string containing the dump
	 */
	virtual char * Dump() { return dump; }
	/**
	 * Retrieves whether the feature has changed at the last frame. This method has to be reimplemented by subclasses.
	 *
	 * @return true if feature has changed, false if not
	 */
	virtual bool Changed() { return changed; }
	/**
	 * Retrieves the feature's type.
	 *
	 * @return type of the feature (0 = none (default), 1 = PES, 2 = MPEG audio, 3 = MPEG video, 4 = AC3 audio, 5 = raw audio, 6 = raw video)
	 */
	virtual int Type() { return type; }
protected:
	int type;
	bool changed;
	static char dump[256];
};

typedef struct
{
	int top;
	int bottom;
	int left;
	int right;
} tBorder;

/**
 * A class representing border detection. It detects the black border a video frame has.
 */
class cBorderFeature
:	public cFeature
{
public:
	/**
	 * Constructs the feature. Initializes the feature's data.
	 */
	cBorderFeature();
	/**
	 * Destructs the feature. Actually does nothing.
	 */
	virtual ~cBorderFeature();
	/**
	 * Refreshs the feature's data.
	 * First the luminance of the picture is calculated. If it is greater than
	 * kMinPictureLuminance, the borders (top, bottom, left, right) will be
	 * calculated. Then the new border is compared with the old border. If it
	 * has changed, the feature's state is set to "changed".
	 *
	 * @param frame video frame which border will be calculated.
	 */
	virtual void Refresh(void * frame);
	/**
	 * Dumps the feature's data (luminance, top, bottom, left, right border,
	 * luminance of first and last lines and columns) to a string.
	 *
	 * @return null-terminated string containing the feature's data.
	 */
	virtual char * Dump();
	/**
	 * Retrieves the border of the last refreshed video frame.
	 *
	 * @return border of the last refreshed video frame.
	 */
	tBorder Border() { return newBorder; }
	/**
	 * Retrieves the luminance of the last refreshed video frame.
	 *
	 * @return luminance of the last refreshed video frame.
	 */
	int GetLuminance() { return luminance; }
private:
	/**
	 * Searches the horizontal borders of the video frame. The luminance and
	 * the ratio between black (luminance < 21) and non-black (luminance >= 21)
	 * pixels are calculated per line. A border will be recognized, if the
	 * luminance of a line is more than twice as high as of the previous line
	 * or the ratio of black/non-black pixels of a line is half as high as of
	 * the previous line. If the ratio is lower than 0.7 the search is ended
	 * with no result.
	 *
	 * @param top controls if the top or bottom border has to be recognized
	 * @param start line where the search starts (0 for top, height - 1 for bottom)
	 */
	void CheckHorizontal(bool top, int start);
	/**
	 * Searches the vertical borders of the video frame. The luminance and
	 * the ratio between black (luminance < 21) and non-black (luminance >= 21)
	 * pixels are calculated per column. A border will be recognized, if the
	 * luminance of a column is more than twice as high as of the previous column
	 * or the ratio of black/non-black pixels of a column is half as high as of
	 * the previous column If the ratio is lower than 0.7 the search is ended
	 * with no result.
	 *
	 * @param left controls if the left or right border has to be recognized
	 * @param start column where the search starts (0 for left, width - 1 for right)
	 */
	void CheckVertical(bool left, int start);
	/**
	 * Calculates the luminance of the video frame. The luminances of
	 * all pixels in the region (20, 150) - (width - 20, height - 150) are
	 * summarized and divided through the number of pixels the region has.
	 *
	 * @return luminance of the video frame.
	 */
	int CheckLuminance();
	cRawVideoFrame * frame;
	tBorder oldBorder;
	tBorder newBorder;
	int firstLine;
	int lastLine;
	int firstCol;
	int lastCol;
	int luminance;

	static const int kThresholdBinarization;
	static const int kMaxHorz;
	static const int kMaxVert;
	static const int kMaxLineLuminance;
	static const int kMinPictureLuminance;
	static const int kFactor;
};

typedef struct
{
	int x;
	int y;
} tPosition;

/**
 * A class representing a station logo. It consists of a name, a station ID,
 * a edge template, a luminance mask and the positions, where the logo can
 * occur.
 */
class cLogoTemplate
{
public:
	/**
	 * Constructs the logo template. Sets the name, station ID. Loads the
	 * edge-template and the luminance mask. Calculates the normal factor.
	 *
	 * @param name null-terminated string containing the name of the logo
	 * @param station station ID of the logo
	 * @param edgefile null-terminated string containing the filename of the PGM file containing the edge template
	 * @param maskfile null-terminated string containing the filename of the PGM file containing the luminance mask
	 */
	cLogoTemplate(char * name, int station, char * edgefile, char * maskfile);
	/**
	 * Destructs the logo template. Actually does nothing.
	 */
	~cLogoTemplate();
	/**
	 * Adds a position where the logo can occur.
	 *
	 * @param x x-coordinate
	 * @param y y-coordinate
	 */
	void AddPosition(int x, int y);
	/**
	 * Retrieves the name of the logo.
	 *
	 * @return null-terminated string containing the name of the logo
	 */
	char * Name() { return name; }
	/**
	 * Retrieves the station id of the logo.
	 *
	 * @return station ID
	 */
	int Station() { return station; }
	/**
	 * Retrieves the width of the logo template.
	 *
	 * @return width of the logo template
	 */
	int Width() { return width; }
	/**
	 * Retrieves the height of the logo template.
	 *
	 * @return height of the logo template
	 */
	int Height() { return height; }
	/**
	 * Retrieves the normal factor of the logo template. The normal factor is
	 * the number of white pixels the edge-template has.
	 *
	 * @return normal factor of the logo template
	 */
	int NormalFactor() { return normalFactor; }
	/**
	 * Retrieves the possible logo position of the given number.
	 *
	 * @return logo position of the given number
	 */
	tPosition Position(int i) { return positions[i]; }
	/**
	 * Retrieves the number of possible positions the logo has.
	 *
	 * @return number of possible logo positions
	 */
	int PositionCount() { return positions.size(); }
	/**
	 * Retrieves the edge template of the logo.
	 *
	 * @return a pointer to the edge template of the logo
	 */
	unsigned char * Data() { return data; }
	/**
	 * Retrieves the luminance mask of the logo.
	 *
	 * @return a pointer to the luminance mask of the logo
	 */
	unsigned char * Mask() { return mask; }
private:
	/**
	 * Loads a PGM (Portable Gray Map) file.
	 *
	 * @param buffer a pointer to an array of unsigned char where the pixel data of the PGM file will be saved to
	 * @param filename name of the PGM file
	 */
	void LoadPGM(unsigned char * buffer, char * filename);
	/**
	 * Parses the buffer till the end of the current line
	 *
	 * @param buffer a pointer to the PGM file's data
	 * @param pos current position in the PGM file
	 *
	 * @return position where the next line in the PGM file starts
	 */
	int NextLine(unsigned char * buffer, int pos);
	char name[100];
	int station;
	int width;
	int height;
	int normalFactor;
	vector <tPosition> positions;
	unsigned char data[100 * 100];
	unsigned char mask[100 * 100];
};

/**
 * A class representing logo detection. It detects the station logo using a
 * cross-correlation between an edge picture of the current frame and the edge
 * template of the station logo.
 */
class cLogoFeature
:	public cFeature
{
public:
	/**
	 * Constructs the feature. Initializes the feature's data. Loads the
	 * different logos of different stations.
	 */
	cLogoFeature();
	/**
	 * Destructs the feature.
	 */
	virtual ~cLogoFeature();
	/**
	 * Refreshs the feature's data. The logo is searched and if found new logo
	 * index and position will be set. If no logo is found logo index is set to
	 * -1. Then the new logo (index and position) is compared with the old one.
	 * If it has changed, the feature's state is set to "changed".
	 *
	 * @param frame video frame in which the logo will be searched.
	 */
	virtual void Refresh(void * frame);
	/**
	 * Dumps the feature's data (logo name, logo position, luminance of the
	 * edge and the frame mask) to a string.
	 *
	 * @return null-terminated string containing the feature's data.
	 */
	virtual char * Dump();
	/**
	 * Retrieves the logo index of the last refreshed frame.
	 *
	 * @return logo index of the last refreshed frame
	 */
	int LogoIndex() { return newLogoIndex; }
	/**
	 * Retrieves the logo position of the last refreshed frame.
	 *
	 * @return logo position of the last refreshed frame
	 */
	tPosition Position() { return newPos; }
private:
	/**
	 * Main logic for logo detection.
	 * If a logo was found at the last frame, the logo is searched at its
	 * last position. If it is found again, the luminance against the edge
	 * picture will be calculated at the current position. It has to be lower
	 * than kMaxEdgeMaskLuminance to be accepted. Then the luminance against
	 * the video frame is calculated at the current position. It has to be
	 * higher than kMaxNoLogoMaskLuminance to be accepted.
	 * If the logo is not found again, the luminance will be checked at the
	 * current position. It it is higher than kMinLogoMaskLuminance it will be
	 * accepted nevertheless, otherwise logo index is set to -1 which stands
	 * for "no logo found".
	 *
	 * If no logo was found at the last frame, all logos will be searched at
	 * their possible positions. If a logo was found in the past only logos
	 * with the remembered station ID are used. If a logo is found again, the
	 * luminance against the edge picture will be calculated at the current
	 * position. It has to be lower than kMaxEdgeMaskLuminance to be accepted.
	 * Then the luminance against the video frame is calculated at the current
	 * position. It has to be higher than kMaxNoLogoMaskLuminance to be
	 * accepted.
	 * If no logo is found again, the logo index is set to -1 which stands for
	 * "no logo found".
	 */
	void CheckLogo();
	/**
	 * Calculates the edge picture of the video frame using a horizontal and
	 * vertical Sobel operator. To speed this up only the first and last 150
	 * lines are calculated.
	 *
	 * @param picture luminance component of the video frame
	 * @param lineSize size of one line in bytes
	 * @param width width of the video frame
	 * @param height height of the video frame
	 */
	void CalcEdges(unsigned char * picture, int lineSize, int width, int height);
	/**
	 * Searches the given logo at its possible positions using a
	 * cross-correlation between the edge picture and the edge template of the
	 * station logo. The maximum value in the output buffer is searched and its
	 * position. If the value is greater than kThresholdLogoMatching a logo has
	 * been found and true is returned.
	 *
	 * @param lineSize size of one line of the edge picture in pixels
	 * @param width width of the edge picture
	 * @param height height of the edge picture
	 * @param logo logo to search for
	 * @param retPos returned position of the maximum value
	 * @param retmaxVal returned maximum value
	 *
	 * @return true if a logo is found, false if not
	 */
	bool MatchLogo(int lineSize, int width, int height,
				   cLogoTemplate *logo,
				   tPosition *retPos, int *retMaxVal);
	/**
	 * Searches the given logo at the given position using a cross-correlation
	 * between the edge picture and the edge template of the station logo. The
	 * maximum value in the output buffer is searched and its position. If the
	 * value is greater than kThresholdLogoMatching a logo has been found and
	 * true is returned.
	 *
	 * @param lineSize size of one line of the edge picture in pixels
	 * @param width width of the edge picture
	 * @param height height of the edge picture
	 * @param logo logo to search for
	 * @param retPos position where to search and returned position of the maximum value
	 * @param retmaxVal returned maximum value
	 *
	 * @return true if a logo is found, false if not
	 */
	bool MatchLogoPosition(int lineSize, int width, int height,
						   cLogoTemplate *logo,
						   tPosition *retPos, int *retMaxVal);
	/**
	 * Calculates the average luminace of the given picture using the luminance
	 * mask of the given logo at the given position.
	 *
	 * @param picture picture for luminance check
	 * @param lineSize size of one line of picture in pixels
	 * @param width width of the picture
	 * @param height height of the picture
	 * @param logo logo which luminance mask is used
	 * @param logoPos position where to calculate the luminance
	 *
	 * @return average luminance of the luminance mask at the given position
	 */
	int MaskLuminance(unsigned char * picture, int lineSize,
					  int width, int height, cLogoTemplate *logo,
					  tPosition logoPos);
	cRawVideoFrame * frame;
	tPosition oldPos;
	tPosition newPos;
	int oldLogoIndex;
	int newLogoIndex;
	int edgeMaskLuminance;
	int logoMaskLuminance;
	int station;
	unsigned char edgePicture[512 * 1024];
	unsigned char outbuf[512 * 1024];
	vector <cLogoTemplate *> logos;

	static const int kThresholdBinarization;
	static const int kThresholdLogoMatching;
	static const int kMinLogoMaskLuminance;
	static const int kMaxEdgeMaskLuminance;
	static const int kMaxNoLogoMaskLuminance;
	static const int kBorder;
};

/**
 * A class representing detection of black-and-white programmes.
 */
class cBlackWhiteFeature
:	public cFeature
{
public:
	/**
	 * Constructs the feature. Initializes the feature's data.
	 */
	cBlackWhiteFeature();
	/**
	 * Destructs the feature. Actually does nothing.
	 */
	virtual ~cBlackWhiteFeature();
	/**
	 * Refreshs the feature's data. First the black-and-white state is
	 * checked. Then the new state is compared with the old one. If it has
	 * changed, the feature's state is set to "changed".
	 *
	 * @param frame video frame which black-and-white state will be checked
	 */
	virtual void Refresh(void * frame);
	/**
	 * Dumps the feature's data (black-and-white state) to a string.
	 *
	 * @return null-terminated string containing the feature's data.
	 */
	virtual char * Dump();
	/**
	 * Retrieves the black-and-white state of the last refreshed frame.
	 *
	 * @return black-and-white state of the last refreshed frame (0 = black-and-white, 1 = colour)
	 */
	int State() { return newState; }
private:
	/**
	 * Checks the black-and-white state of the current frame. Therefore the
	 * region (10, 70) - (width - 1 - 10, height - 1 - 70) of the chominance
	 * components is checked for pixels which value is between kMinChrominance
	 * and kMaxChrominance. If the ratio of such pixels and all pixels of the
	 * region is < 0.90 then the frame is a colour frame else the frame is a
	 * black-and-white frame.
	 */
	void Check();
	cRawVideoFrame * frame;
	int oldState;
	int newState;

	static const int kBlackWhite;
	static const int kColour;
	static const int kMinChrominance;
	static const int kMaxChrominance;
};

/**
 * A class representing detection of changes in AC3 channel modes.
 */
class cAC3ChannelFeature
:	public cFeature
{
public:
	/**
	 * Constructs the feature. Initializes the feature's data.
	 */
	cAC3ChannelFeature();
	/**
	 * Destructs the feature. Actually does nothing.
	 */
	virtual ~cAC3ChannelFeature();
	/**
	 * Refreshs the feature's data. First the channel mode of the current AC3
	 * audio frame is checked. Then the new channel mode is compared with the
	 * old one. If it has changed, the feature's state is set to "changed".
	 *
	 * @param frame AC3 audio frame which channel mode will be checked.
	 */
	virtual void Refresh(void * frame);
	/**
	 * Dumps the feature's data (channel mode) to a string.
	 *
	 * @return null-terminated string containing the feature's data.
	 */
	virtual char * Dump();
	/**
	 * Retrieves the channel mode of the last refreshed frame.
	 *
	 * @return channel mode of the last refreshed frame
	 */
	int ChannelMode() { return newChannelMode; }
private:
	/**
	 * Checks the channel mode of the current frame.
	 */
	void CheckChannels();
	cAC3AudioFrame * frame;
	int oldChannelMode;
	int newChannelMode;
};

/**
 * A class representing detection of mono programmes, especially mono
 * programmes sended as stereo.
 */
class cMonoStereoFeature
:	public cFeature
{
public:
	/**
	 * Constructs the feature. Initializes the feature's data.
	 */
	cMonoStereoFeature();
	/**
	 * Destructs the feature. Actually does nothing.
	 */
	virtual ~cMonoStereoFeature();
	/**
	 * Refreshs the feature's data. First the channels of the current audio
	 * frame is checked. Then the new channels is compared with the old one.
	 * If it has changed, the feature's state is set to "changed".
	 *
	 * @param frame audio frame which channels will be checked.
	 */
	virtual void Refresh(void * frame);
	/**
	 * Dumps the feature's data (channels) to a string.
	 *
	 * @return null-terminated string containing the feature's data.
	 */
	virtual char * Dump();
	/**
	 * Retrieves the channels of the last refreshed frame.
	 *
	 * @return channels of the last refreshed frame (1 = mono, 2 = stereo)
	 */
	int Channels() { return newChannels; }
private:
	/**
	 * Checks the channels of the current frame. The channel mode of the MPEG
	 * audio frame is checked. If it is stereo all samples of the audio frame
	 * will be checked to determine if the frame contains really a stereo
	 * signal and not a mono signal sended as stereo.
	 */
	void CheckChannels();
	cRawAudioFrame * frame;
	int oldChannels;
	int newChannels;
};

#endif
