/*
 * pes.cpp: PES frame demuxing
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

extern "C"
{
#include <ffmpeg/avcodec.h>
}

#include "pes.h"
#include "util.h"
#include "tools.h"
#include "audio.h"

const unsigned int kStreamTypeUnknown = 0;
const unsigned int kStreamTypeAudio   = 1;
const unsigned int kStreamTypeVideo   = 2;
const unsigned int kStreamTypePrivate = 3;

unsigned char audioStreamRest[8][2048];
unsigned int audioStreamRestFrameSize[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
unsigned int audioStreamRestLength[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
unsigned char AC3StreamRest[2048];
unsigned int AC3StreamRestFrameSize = 0;
unsigned int AC3StreamRestLength = 0;

// ------------- cMPEGStream -----------------------------------------

cMPEGStream::cMPEGStream(unsigned char id, unsigned char * data, unsigned int length)
:	cBitStream(data, length),
	id(id),
	frameCount(0),
	rest(false),
	copyRest(false)
{
	switch (this->id & 0xF0)
	{
	case 0xB0:
		this->type = kStreamTypePrivate;
		break;
	case 0xC0:
		this->type = kStreamTypeAudio;
		this->data[this->length] = 0xFF;
		break;
	case 0xE0:
		this->type = kStreamTypeVideo;
		this->frameCount = 1;
		this->data[this->length] = 0x00;
		this->data[this->length + 1] = 0x00;
		this->data[this->length + 2] = 0x01;
		this->data[this->length + 3] = 0xB3;
		break;
	default:
		this->type = kStreamTypeUnknown;
		break;
	}
	Parse();
}

cMPEGStream::~cMPEGStream()
{
	if (copyRest)
	{
		if (this->type == kStreamTypePrivate)
		{
			memcpy(AC3StreamRest,
				   &this->data[frameOffsets[frameCount]],
				   AC3StreamRestLength);
		}
		else
		{
			memcpy(audioStreamRest[this->id & 0x0F],
				   &this->data[frameOffsets[frameCount]],
				   audioStreamRestLength[this->id & 0x0F]);
		}
	}
	if (this->length > 0)
		delete[] this->data;
}

void cMPEGStream::Parse()
{
	if (this->type == kStreamTypePrivate)
	{
		unsigned long header;
		int frameSize = 0;

		header = NextBits(16);
		if ((header & 0xFFFF) != 0x0B77)
		{
			GetBits(8);
			while (!this->IsOnEnd())
			{
				header = NextBits(16);
				if ((header & 0xFFFF) == 0x0B77)
				{
					if (AC3HeaderValid())
						break;
				}
				GetBits(8);
			}
			if (!this->IsOnEnd())
			{
				memcpy(&AC3StreamRest[AC3StreamRestLength],
					   this->data,
					   this->byteOffset);
				AC3StreamRestLength += this->byteOffset;
			}
			else
			{
				memcpy(&AC3StreamRest[AC3StreamRestLength],
					   this->data,
					   this->length);
				AC3StreamRestLength += this->length;
			}
			frameSize = AC3StreamRestFrameSize;
			if (frameSize == (int) AC3StreamRestLength)
			{
				this->frameOffsets[0] = AC3StreamRestLength;
				this->frameCount = 1;
				AC3StreamRestLength = 0;
				AC3StreamRestFrameSize = 0;
				rest = true;
			}
			else
			{
				AC3StreamRestLength = 0;
				AC3StreamRestFrameSize = 0;
			}
		}
		if (!this->IsOnEnd())
		{
			while (!this->IsOnEnd())
			{
				header = NextBits(16);
				if ((header & 0xFFFF) != 0x0B77)
					break;
				frameSize = this->CalcAC3FrameSize();
				frameOffsets[frameCount] = byteOffset;
				frameCount++;
				SeekBytes(frameSize);
			}
			frameOffsets[frameCount] = byteOffset + 1;
			int restBytes = frameOffsets[frameCount] - frameOffsets[frameCount - 1];
			if (restBytes != frameSize)
			{
				this->copyRest = true;
				frameCount--;
				AC3StreamRestFrameSize = frameSize;
				AC3StreamRestLength = restBytes;
			}
		}
	}
	if (this->type == kStreamTypeAudio)
	{
		unsigned long header;
		int frameSize = 0;

		header = NextBits(16);
		if ((header & 0xFFE0) != 0xFFE0)
		{
			GetBits(8);
			while (!this->IsOnEnd())
			{
				header = NextBits(16);
				if ((header & 0xFFE0) == 0xFFE0)
					if (AudioHeaderValid(NextBits(32)))
						break;
				GetBits(8);
			}
			int num = this->id & 0x0F;
			if (!this->IsOnEnd())
			{
				memcpy(&audioStreamRest[num][audioStreamRestLength[num]],
					   this->data,
					   this->byteOffset);
				audioStreamRestLength[num] += this->byteOffset;
			}
			else
			{
				memcpy(&audioStreamRest[num][audioStreamRestLength[num]],
					   this->data,
					   this->length);
				audioStreamRestLength[num] += this->length;
			}
			frameSize = audioStreamRestFrameSize[num];
			if (frameSize == (int) audioStreamRestLength[num])
			{
				this->frameOffsets[0] = audioStreamRestLength[num];
				this->frameCount = 1;
				audioStreamRestLength[num] = 0;
				audioStreamRestFrameSize[num] = 0;
				rest = true;
			}
			else
			{
				audioStreamRestLength[num] = 0;
				audioStreamRestFrameSize[num] = 0;
			}
		}
		if (!this->IsOnEnd())
		{
			while (!this->IsOnEnd())
			{
				header = NextBits(32);
				if ((header & 0xFFE00000) != 0xFFE00000)
					break;
				frameSize = this->CalcFrameSize(header);
				frameOffsets[frameCount] = byteOffset;
				frameCount++;
				SeekBytes(frameSize);
			}
			frameOffsets[frameCount] = byteOffset + 1;
			int restBytes = frameOffsets[frameCount] - frameOffsets[frameCount - 1];
			if (restBytes != frameSize)
			{
				this->copyRest = true;
				frameCount--;
				audioStreamRestFrameSize[this->id & 0x0F] = frameSize;
				audioStreamRestLength[this->id & 0x0F] = restBytes;
			}
		}
	}
}

bool cMPEGStream::AudioHeaderValid(unsigned long header)
{
	unsigned long i;
	// Version
	i = (header >> 19) & 0x03;
	if (i == 1)
		return false;
	// Layer
	i = (header >> 17) & 0x03;
	if (i == 0)
		return false;
	// Bitrate
	i = (header >> 12) & 0x0F;
	if (i == 0 || i == 15)
		return false;
	// Samplerate
	i = (header >> 10) & 0x03;
	if (i == 3)
		return false;

	return true;
}

int cMPEGStream::CalcFrameSize(unsigned long header)
{
	int lsf;
	int mpeg25;
	int layer;
	int samplingRate;
	int bitRate;
	int padding;
	int frameSize = 0;

	if (header & (1 << 20))
	{
		lsf = (header & (1 << 19)) ? 0 : 1;
		mpeg25 = 0;
	}
	else
	{
		lsf = 1;
		mpeg25 = 1;
	}
	layer = 4 - ((header >> 17) & 3);
	samplingRate = ((header >> 10) & 3);
	samplingRate = kFreqTab[samplingRate] >> (lsf + mpeg25);

	bitRate = (header >> 12) & 0xF;
	bitRate = kBitrateTab[lsf][layer - 1][bitRate];
	padding = (header >> 9) & 0x1;
	switch(layer)
	{
	case 1:
		frameSize = ((bitRate * 12000) / samplingRate + padding) << 2;
		break;
	case 2:
		frameSize = (bitRate * 144000) / samplingRate + padding;
		break;
	case 3:
		frameSize = (bitRate * 144000) / (samplingRate << lsf) + padding;
		break;
	}
	return frameSize;
}

bool cMPEGStream::AC3HeaderValid()
{
	unsigned long offset = byteOffset;
	GetBits(32);
	int frameSizeCode = GetBits(8) & 63;
	int bsid = GetBits(8);
	byteOffset = offset;

	if (frameSizeCode > 38)
		return false;
	if (bsid > 0x60)
		return false;

	return true;
}

int cMPEGStream::CalcAC3FrameSize()
{
	static int rate[] = { 32,  40,  48,  56,  64,  80,  96, 112,
						 128, 160, 192, 224, 256, 320, 384, 448,
						 512, 576, 640};
	int frmsizecod;
	int bitrate;
	unsigned long offset = byteOffset;
	GetBits(32);
	int buf4 = GetBits(8);
	byteOffset = offset;

	frmsizecod = buf4 & 63;
	bitrate = rate [frmsizecod >> 1];

	switch (buf4 & 0xc0)
	{
	case 0:
		return 4 * bitrate;
	case 0x40:
		return 2 * (320 * bitrate / 147 + (frmsizecod & 1));
	case 0x80:
		return 6 * bitrate;
	default:
		return 0;
	}
}

cMPEGAudioFrame * cMPEGStream::GetAudioFrame(int frameNumber)
{
	if (this->type == kStreamTypeAudio)
	{
		if (frameNumber < this->frameCount)
		{
			if (rest && frameNumber == 0)
			{
				cMPEGAudioFrame *frame;
				frame = new cMPEGAudioFrame(&audioStreamRest[this->id & 0x0F][0],
											this->frameOffsets[0]);
				return frame;
			}
			else
			{
				int start = this->frameOffsets[frameNumber];
				int end = this->frameOffsets[frameNumber + 1];
				cMPEGAudioFrame *frame;
				frame = new cMPEGAudioFrame(&this->data[start], end - start);
				return frame;
			}
		}
	}
	return NULL;
}

cAC3AudioFrame * cMPEGStream::GetAC3AudioFrame(int frameNumber)
{
	if (this->type == kStreamTypePrivate)
	{
		if (frameNumber < this->frameCount)
		{
			if (rest && frameNumber == 0)
			{
				cAC3AudioFrame *frame;
				frame = new cAC3AudioFrame(&AC3StreamRest[0],
											this->frameOffsets[0]);
				return frame;
			}
			else
			{
				int start = this->frameOffsets[frameNumber];
				int end = this->frameOffsets[frameNumber + 1];
				cAC3AudioFrame *frame;
				frame = new cAC3AudioFrame(&this->data[start], end - start);
				return frame;
			}
		}
	}
	return NULL;
}

cMPEGVideoFrame * cMPEGStream::GetVideoFrame(int frameType)
{
	if (this->type == kStreamTypeVideo)
	{
		cMPEGVideoFrame * frame;
 		frame = new cMPEGVideoFrame(this->data, this->length + 4, frameType);
		return frame;
	}
	return NULL;
}


// ------------- cPESFrame -------------------------------------------

cPESFrame::cPESFrame(unsigned char * data, unsigned int length, int frameType)
:	cBitStream(data, length),
	frameType(frameType),
	streamCount(0),
	audioStreamCount(0),
	videoStreamCount(0),
	privateStreamCount(0)
{
	Parse();
}

cPESFrame::~cPESFrame()
{
	for (int i = 0; i < this->streamCount; i++)
		delete this->streams[i];
}

void cPESFrame::Parse()
{
	unsigned char streamID[16];
	unsigned int streamLength[16];
	unsigned char * streamData[16];
	unsigned int streamDataLength[16];
	unsigned long long streamPTS[16];
	unsigned char id;
	int i;
	unsigned int packetLength;
	unsigned int headerDataLength;
	unsigned char ptsdtsflag;
	unsigned int offset;

	for (i = 0; i < 16; i++)
		streamLength[i] = 0;
	while (!this->IsOnEnd())
	{
		this->SeekBytes(3);
		id = this->GetBits(8);
		packetLength = this->GetBits(16);
		if (id & 0xF0 != 0xC0 &&
			id & 0xF0 != 0xE0 &&
			id != 0xBD)
		{
			this->SeekBytes(packetLength);
			continue;
		}
		this->SeekBytes(1);
		ptsdtsflag = this->GetBits(2);
		this->AlignBits();
		headerDataLength = this->GetBits(8);
		for (i = 0; i < this->streamCount; i++)
			if (streamID[i] == id)
				break;
		if (i == this->streamCount)
		{
			streamID[this->streamCount] = id;
			this->streamCount++;
			switch(id & 0xF0)
			{
			case 0xB0:
				this->privateStreamCount++;
				break;
			case 0xC0:
				this->audioStreamCount++;
				break;
			case 0xE0:
				this->videoStreamCount++;
				break;
			}
		}
		if (ptsdtsflag == 2 || ptsdtsflag == 3)
		{
			this->SeekBits(4);
			streamPTS[i] = this->GetBits(3) << 15;
			this->SeekBits(1);
			streamPTS[i] |= this->GetBits(15) << 15;
			this->SeekBits(1);
			streamPTS[i] |= this->GetBits(15);
			this->AlignBits();
			offset = 5;
		}
		else
		{
			offset = 0;
		}
		streamLength[i] += packetLength - 3 - headerDataLength;
		this->SeekBytes(packetLength - 3 - offset);
	}
	this->SeekStart();
	for (i = 0; i < this->streamCount; i++)
	{
		streamData[i] = new unsigned char[streamLength[i] + 10];
		streamDataLength[i] = 0;
	}
	while (!this->IsOnEnd())
	{
		this->SeekBytes(3);
		id = this->GetBits(8);
		packetLength = this->GetBits(16);
		if (id & 0xF0 != 0xC0 &&
			id & 0xF0 != 0xE0 &&
			id != 0xBD)
		{
			this->SeekBytes(packetLength);
			continue;
		}
		this->SeekBytes(2);
		headerDataLength = this->GetBits(8);
		for (i = 0; i < this->streamCount; i++)
			if (streamID[i] == id)
				break;
		memcpy(&streamData[i][streamDataLength[i]],
			   &data[byteOffset + headerDataLength],
			   packetLength - 3 - headerDataLength);
		streamDataLength[i] += packetLength - 3 - headerDataLength;
		this->SeekBytes(packetLength - 3);
	}
	for (i = 0; i < this->streamCount; i++)
	{
		this->streams[i] = new cMPEGStream(streamID[i],
										   streamData[i],
										   streamLength[i]);
		this->streams[i]->SetPTS(streamPTS[i]);
	}
}

unsigned char cPESFrame::GetStreamID(int streamNumber)
{
	if (streamNumber < streamCount)
	{
		return streams[streamNumber]->GetID();
	}
	return 0;
}

int cPESFrame::GetAudioFrameCount(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->audioStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypeAudio)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetFrameCount();
				}
				num++;
			}
		}
	}
	return 0;
}

int cPESFrame::GetAC3AudioFrameCount(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->privateStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypePrivate)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetFrameCount();
				}
				num++;
			}
		}
	}
	return 0;
}

cMPEGAudioFrame * cPESFrame::GetAudioFrame(int streamNumber, int frameNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->audioStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypeAudio)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetAudioFrame(frameNumber);
				}
				num++;
			}
		}
	}
	return NULL;
}

cAC3AudioFrame * cPESFrame::GetAC3AudioFrame(int streamNumber, int frameNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->privateStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypePrivate)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetAC3AudioFrame(frameNumber);
				}
				num++;
			}
		}
	}
	return NULL;
}

cMPEGVideoFrame * cPESFrame::GetVideoFrame(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->videoStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypeVideo)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetVideoFrame(this->frameType);
				}
				num++;
			}
		}
	}
	return NULL;
}

unsigned long long cPESFrame::GetAudioPTS(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->audioStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypeAudio)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetPTS();
				}
				num++;
			}
		}
	}
	return 0;
}

unsigned long long cPESFrame::GetPrivatePTS(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->privateStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypePrivate)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetPTS();
				}
				num++;
			}
		}
	}
	return 0;
}

unsigned long long cPESFrame::GetVideoPTS(int streamNumber)
{
	int i;
	int num;
	
	if (streamNumber < this->videoStreamCount)
	{
		num = 0;
		for (i = 0; i < this->streamCount; i++)
		{
			if (this->streams[i]->GetType() == kStreamTypeVideo)
			{
				if (num == streamNumber)
				{
					return this->streams[i]->GetPTS();
				}
				num++;
			}
		}
	}
	return 0;
}

