/*
 * pes.h: PES frame demuxing
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#ifndef _PES_H_
#define _PES_H_

#include "stream.h"
#include "audio.h"
#include "video.h"

/**
 * A class representing a MPEG stream of one PES frame.
 * This can be audio, video and private (AC3 audio) streams.
 * It is a helper class used by cPESFrame for demuxing.
 */
class cMPEGStream
: public cBitStream
{
public:
	/**
	 * Constructs the MPEG stream.
	 *
	 * @param id stream ID
	 * @param data pointer to the stream data
	 * @param length length of the stream data
	 */
	cMPEGStream(unsigned char id, unsigned char * data, unsigned int length);
	/**
	 * Destructs the MPEG stream.
	 */
	~cMPEGStream();
	/**
	 * Sets the PTS of the stream. PTS stands for Presentation Time Stamp.
	 *
	 * @param pts new PTS
	 */
	void SetPTS(unsigned long long pts) { this->pts = pts; }
	/**
	 * Retrieves the PTS of the stream. PTS stands for Presentation Time Stamp.
	 *
	 * @return the PTS of the stream
	 */
	unsigned long long GetPTS() { return this->pts; }
	/**
	 * Retrieves the stream ID.
	 *
	 * @return stream ID
	 */
	unsigned char GetID() { return this->id; }
	/**
	 * Retrieves the type of the stream.
	 *
	 * @return the type of the stream (0 = Unknown, 1 = Audio, 2 = Video, 3 = Private)
	 */
	unsigned int GetType() { return this->type; }
	/**
	 * Retrieves the number of frames the stream contains. For video streams number of frames is always 1.
	 *
	 * @return number of frames
	 */
	int GetFrameCount() { return this->frameCount; }
	/**
	 * Retrieves the MPEG audio frame of the given number, according to the internal frame offset table.
	 *
	 * @param frameNumber number of the frame to be retrieved (0 to (frame count - 1)). If no frame of this number exists, NULL will be returned.
	 *
	 * @return the retrieved MPEG audio frame, or NULL if an error occured
	 */
	cMPEGAudioFrame * GetAudioFrame(int frameNumber);
	/**
	 * Retrieves the AC3 audio frame of the given number, according to the internal frame offset table.
	 *
	 * @param frameNumber number of the frame to be retrieved (0 to (frame count - 1)). If no frame of this number exists, NULL will be returned.
	 *
	 * @return the retrieved AC3 audio frame, or NULL if an error occured
	 */
	cAC3AudioFrame * GetAC3AudioFrame(int frameNumber);
	/**
	 * Retrieves the MPEG video frame of the stream.
	 *
	 * @param frameType type of the video frame (I, P, B, D)
	 *
	 * @return the retrieved MPEG video frame, or NULL if an error occured
	 */
	cMPEGVideoFrame * GetVideoFrame(int frameType);
private:
	/**
	 * Parses the stream to fill the frame offset table.
	 */
	void Parse();
	/**
	 * Checks whether the given MPEG audio header is valid.
	 *
	 * @param header the MPEG audio header to check
	 *
	 * @return true if header is valid, false if not
	 */
	bool AudioHeaderValid(unsigned long header);
	/**
	 * Calculates the frame size of the MPEG audio frame that is starting with the given header.
	 *
	 * @param header the MPEG audio header of the frame
	 *
	 * @return frame size in bytes
	 */
	int CalcFrameSize(unsigned long header);
	/**
	 * Checks whether the following AC3 audio header is valid. The next bytes of the stream are used.
	 *
	 * @return true if header is valid, false if not
	 */
	bool AC3HeaderValid();
	/**
	 * Calculates the frame size of the next AC3 audio frame. The next bytes of the stream are used.
	 *
	 * @return frame size in bytes
	 */
	int CalcAC3FrameSize();
	unsigned char id;
	unsigned int type;
	unsigned long long pts;
	int frameOffsets[100];
	int frameCount;
	bool rest;
	bool copyRest;
};

/**
 * A class representing a MPEG PES frame. PES stands for Packetized Elementary Stream.
 */
class cPESFrame
: public cBitStream
{
public:
	/**
	 * Constructs the PES frame.
	 * The data is parsed, the streams are demuxed and corresponding cMPEGStream objects are created.
	 *
	 * @param data pointer to the stream data
	 * @param length length of the stream data
	 * @param frameType type of the video frame the PES frame contains
	 */
	cPESFrame(unsigned char * data, unsigned int length, int frameType);
	/**
	 * Destructs the PES frame.
	 */
	~cPESFrame();
	/**
	 * Retrieves the number of streams the PES frame contains.
	 * This includes audio, video and private streams.
	 *
	 * @return number of streams
	 */
	int GetStreamCount() { return this->streamCount; }
	/**
	 * Retrieves the number of audio streams the PES frame contains.
	 *
	 * @return number of audio streams
	 */
	int GetAudioStreamCount() { return this->audioStreamCount; }
	/**
	 * Retrieves the number of video streams the PES frame contains.
	 *
	 * @return number of video streams
	 */
	int GetVideoStreamCount() { return this->videoStreamCount; }
	/**
	 * Retrieves the number of private streams the PES frame contains.
	 *
	 * @return number of private streams
	 */
	int GetPrivateStreamCount() { return this->privateStreamCount; }
	/**
	 * Retrieves the ID of the stream of the given number.
	 *
	 * @param streamNumber number of the stream (0 to (stream count - 1)). If no stream of this number exists, 0 will be returned.
	 *
	 * @return the retrieved stream ID, or 0 if streamNumber is invalid
	 */
	unsigned char GetStreamID(int streamNumber);
	/**
	 * Retrieves the number of frames the audio stream of the given number contains.
	 *
	 * @param streamNumber number of the audio stream which frame count has to be retrieved. If no stream of this number exists, 0 will be returned.
	 *
	 * @return number of frames the audio stream contains, or 0 if streamNumber is invalid
	 */
	int GetAudioFrameCount(int streamNumber);
	/**
	 * Retrieves the number of frames the AC3 audio stream of the given number contains.
	 *
	 * @param streamNumber number of the AC3 audio stream which frame count has to be retrieved. If no stream of this number exists, 0 will be returned.
	 *
	 * @return number of frames the AC3 audio stream contains, or 0 if streamNumber is invalid
	 */
	int GetAC3AudioFrameCount(int streamNumber);
	/**
	 * Retrieves the MPEG audio frame of the given number of the given stream.
	 *
	 * @param streamNumber number of the audio stream that contains the frame to be retrieved. If no stream of this number exists, NULL will be returned.
	 * @param frameNumber number of the frame to be retrieved. If no frame of this number exists, NULL will be returned.
	 *
	 * @return the retrieved MPEG audio frame, or NULL if an error occured
	 */
	cMPEGAudioFrame * GetAudioFrame(int streamNumber, int frameNumber);
	/**
	 * Retrieves the AC3 audio frame of the given number of the given stream.
	 *
	 * @param streamNumber number of the private stream that contains the frame to be retrieved. If no stream of this number exists, NULL will be returned.
	 * @param frameNumber number of the frame to be retrieved. If no frame of this number exists, NULL will be returned.
	 *
	 * @return the retrieved AC3 audio frame, or NULL if an error occured
	 */
	cAC3AudioFrame * GetAC3AudioFrame(int streamNumber, int frameNumber);
	/**
	 * Retrieves the MPEG video frame of the given stream.
	 *
	 * @param streamNumber number of the video stream that contains the frame to be retrieved. If no stream of this number exists, NULL will be returned.
	 *
	 * @return the retrieved MPEG video frame, or NULL if an error occured
	 */
	cMPEGVideoFrame * GetVideoFrame(int streamNumber);
	/**
	 * Retrieves the PTS of the given audio stream. PTS stands for Presentation Time Stamp.
	 *
	 * @param streamNumber number of the audio stream that PTS has to be retrieved. If no stream of this number exists, 0 will be returned.
	 *
	 * @return the PTS of the audio stream, or 0 if streamNumber is invalid
	 */
	unsigned long long GetAudioPTS(int streamNumber);
	/**
	 * Retrieves the PTS of the given video stream. PTS stands for Presentation Time Stamp.
	 *
	 * @param streamNumber number of the video stream that PTS has to be retrieved. If no stream of this number exists, 0 will be returned.
	 *
	 * @return the PTS of the video stream, or 0 if streamNumber is invalid
	 */
	unsigned long long GetVideoPTS(int streamNumber);
	/**
	 * Retrieves the PTS of the given private stream. PTS stands for Presentation Time Stamp.
	 *
	 * @param streamNumber number of the private stream that PTS has to be retrieved. If no stream of this number exists, 0 will be returned.
	 *
	 * @return the PTS of the private stream, or 0 if streamNumber is invalid
	 */
	unsigned long long GetPrivatePTS(int streamNumber);
private:
	void Parse();
	int frameType;
	cMPEGStream * streams[16];
	int streamCount;
	int audioStreamCount;
	int videoStreamCount;
	int privateStreamCount;
};

#endif

