/*
 * scanner.h: Main logic and analysis
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#ifndef _SCANNER_H_
#define _SCANNER_H_

#include <list>

#include "recording.h"
#include "tools.h"
#include "pes.h"
#include "feature.h"
#include <stdio.h>

extern "C" {
#include <ffmpeg/avcodec.h>
}

typedef struct
{
	int index;
	int state;
} tChange;

typedef struct
{
	int index;
	int gopIndex;
	int state;
	tBorder border;
} tBorderChange;

typedef struct
{
	int index;
	int gopIndex;
	int state;
	int logoIndex;
	tPosition logoPosition;
} tLogoChange;

typedef struct
{
	int index;
	int gopIndex;
	int state;
	int bwState;
} tBlackWhiteChange;

typedef struct
{
	int index;
	int gopIndex;
	int state;
	int channelMode;
} tAC3ChannelChange;

typedef struct
{
	int index;
	int gopIndex;
	int state;
	int channels;
} tMonoStereoChange;

/**
 * This class is responsible for reading the recorded data stream using classes
 * from VDR, demultiplexing and decoding it using the frame and stream classes,
 * collecting change data using the feature classes and doing the final
 * analysis for setting the cutting marks.
 */
class cCommercialScanner
{
public:
	/**
	 * Constructs the scanner. Opens the recording and its index file.
	 *
	 * @param fileName null-terminated string containing the name of the recording
	 */
	cCommercialScanner(const char * fileName);
	/**
	 * Destructs the scanner. Closes the recording and its index file.
	 */
	~cCommercialScanner();
	/**
	 * Main Method of the scanner. Reads the recording, demultiplexes and
	 * decodes it, collects the feature changes and does the final analysis
	 * for setting the cutting marks.
	 */
	void Process();
	/**
	 * Dumps the feature changes to logfile and stdout.
	 */
	void DumpChanges();
	/**
	 * Saves the feature changes to files.
	 */
	void SaveChanges(int prefix);
private:
	/**
	 * Refreshs the features that are based on PES frames.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current PES frame
	 */
	void CheckPESFrameFeatures(int index, int gopIndex, void * pesFrame);
	/**
	 * Refreshs the features that are based on MPEG video frames.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current MPEG video frame
	 */
	void CheckMPEGVideoFeatures(int index, int gopIndex, void * mpegVideo);
	/**
	 * Refreshs the features that are based on uncompressed video frames. These
	 * are border, logo and black-and-white detection. Changes within a short
	 * period of time are handled and under special conditions ignored.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current uncompressed video frame
	 */
	void CheckRawVideoFeatures(int index, int gopIndex, void * rawVideo);
	/**
	 * Refreshs the features that are based on MPEG audio frames.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current MPEG audio frame
	 */
	void CheckMPEGAudioFeatures(int index, int gopIndex, void * mpegAudio);
	/**
	 * Refreshs the features that are based on uncompressed audio frames. This
	 * is the mono-stereo feature. Changes within a short period of time are
	 * handled and under special conditions ignored.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current uncompressed audio frame
	 */
	void CheckRawAudioFeatures(int index, int gopIndex, void * rawAudio);
	/**
	 * Refreshs the features that are based on AC3 audio frames. This is the
	 * channel mode feature.
	 *
	 * @param index index of the frame
	 * @param gopIndex GOP index of the frame
	 * @param pesFrame current AC3 audio frame
	 */
	void CheckAC3AudioFeatures(int index, int gopIndex, void * ac3Audio);

	/**
	 * In border detection frames with insymmetries can occur. This methods
	 * tries to dispose them by comparisons with the previous and next frame.
	 */
	void CheckBorderInsymmetries();

	/**
	 * Main logic for commercial detection. 
	 */
	void Analysis();
	/**
	 * Sets the cutting marks in the recording according to the items of the change list. 
	 */
	void SetCuttingMarks();
	/**
	 * Tests whether the changes in the AC3 list are relevant for the analysis.
	 *
	 * @return result of the test
	 */
	bool AC3ChangesRelevant();
	/**
	 * Tests whether the changes in the black and white list are relevant
	 * for the analysis.
	 *
	 * @return result of the test
	 */
	bool BWChangesRelevant();
	/**
	 * Adds a change to the change list.
	 *
	 * @param index index of the frame
	 * @param state new state (1 = Programme, 2 = Preview, 3 = Commercial)
	 */
 	void AddChange(int index, int state);
	/**
	 * Retrieves the number of border changes between start and end frame.
	 *
	 * @param start index of the start frame
	 * @param end index of the end frame
	 *
	 * @return number of border changes between start and end frame
	 */
	int GetBorderChangeCount(int start, int end);
	/**
	 * Retrieves the border change at the given frame.
	 *
	 * @param index index of the frame
	 *
	 * @return border change at the given frame
	 */
	list <tBorderChange>::iterator GetBorderAtIndex(int index);
	/**
	 * Retrieves the border change with the given minimum length between start
	 * and end frame. If no change serves this conditions the last border
	 * change will be returned.
	 *
	 * @param start index of the start frame
	 * @param end index of the end frame
	 * @param minLength minimum length the border change must have
	 *
	 * @return border change tha serves the conditions
	 */
	list <tBorderChange>::iterator GetBorderChunk(int start, int end, int minLength);

	/**
	 * Compares two logo changes (index and position).
	 *
	 * @param logo1 logo change 1
	 * @param logo2 logo change 2
	 *
	 * @return result of the compare (0 = both no logo or logo index and position equal,
	 *								  1 = logo index equal but position changed,
	 *								  2 = logo index changed,
	 *								  3 = change from logo to no logo or no logo to logo)
	 */	
	int CompareLogos(tLogoChange & logo1, tLogoChange & logo2);
	/**
	 * Compares two borders with tolerance.
	 *
	 * @param border1 border 1
	 * @param border2 border 2
	 *
	 * @return result of the compare (0 = border equal,
	 *								  1 = border changed)
	 */	
	int CompareBorders(tBorder & border1, tBorder & border2);

	/**
	 * Refreshs the given feature.
	 *
	 * @param index index of current frame
	 * @param feature feature to refresh
	 * @param frame frame that is used for refresh
	 */
	void RefreshFeature(int index, cFeature *feature, void * frame);
	/**
	 * Checks if the given feature has changed. Then it dumps index and
	 * features data to stdout and logfile.
	 *
	 * @param index index of current frame
	 * @param feature feature to check
	 */
	void CheckFeature(int index, cFeature *feature);

	char fileName[1000];
	cIndexFile * recIndex;
	cFileName * recFile;
	FILE *logFile;
	AVCodec *audioCodec;
	AVCodecContext *audioCodecContext[16];
	unsigned char frameBuffer[256 * 1024];

	cBorderFeature border;
	cLogoFeature logo;
	cBlackWhiteFeature blackwhite;
	cAC3ChannelFeature ac3channel;
	cMonoStereoFeature monostereo;

	list <tBorderChange> borderChanges;
	list <tLogoChange> logoChanges;
	list <tBlackWhiteChange> bwChanges;
	list <tAC3ChannelChange> ac3Changes;
	list <tMonoStereoChange> monostereoChanges;

	list <tChange> changes;
	list <int> marks;
};

#endif
