/*
 * video.cpp: Video frame handling
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

extern "C"
{
#include <ffmpeg/avcodec.h>
}

#include "video.h"
#include "tools.h"

#include <stdlib.h>
#include <stdio.h>

FILE *rawYVideoFile;
FILE *rawUVideoFile;
FILE *rawVVideoFile;
FILE *mpegVideoFile;

cRawVideoFrame::cRawVideoFrame(AVFrame *picture, int width, int height)
:	width(width),
	height(height)
{
	int i;
	int size;
	size = picture->linesize[0] * height;
	this->picture.linesize[0] = picture->linesize[0];
	this->picture.data[0] = new UINT8[size];
	memcpy(this->picture.data[0], picture->data[0], size);
	for (i = 1; i < 3; i++)
	{
		size = picture->linesize[i] * height / 2;
		this->picture.linesize[i] = picture->linesize[i];
		this->picture.data[i] = new UINT8[size];
		memcpy(this->picture.data[i], picture->data[i], size);
	}
	for (int j = 0; j < height; j++)
	{
//		fwrite(&picture->data[0][j * picture->linesize[0]], width, 1, rawYVideoFile);
//		fwrite(&picture->data[1][j * picture->linesize[1]], width / 2, 1, rawUVideoFile);
//		fwrite(&picture->data[2][j * picture->linesize[2]], width / 2, 1, rawVVideoFile);
	}
}

cRawVideoFrame::~cRawVideoFrame()
{
	int i;
	for (i = 0; i < 3; i++)
	{
		delete[] this->picture.data[i];
	}
}

unsigned char * cRawVideoFrame::PictureData(int i)
{
	if (i < 3)
		return this->picture.data[i];
	return NULL;
}

int cRawVideoFrame::LineSize(int i)
{
	if (i < 3)
		return this->picture.linesize[i];
	return 0;
}


const unsigned char kPictureStartCode	= 0x00;
const unsigned char kUserDataStartCode	= 0xB2;
const unsigned char kSequenceHeaderCode = 0xB3;
const unsigned char kSequenceErrorCode	= 0xB4;
const unsigned char kExtensionStartCode = 0xB5;
const unsigned char kSequenceEndCode 	= 0xB7;
const unsigned char kGOPStartCode		= 0xB8;

const unsigned char kSequenceExtensionID				= 0x01;
const unsigned char kSequenceDisplayExtensionID			= 0x02;
const unsigned char kQuantMatrixExtensionID				= 0x03;
const unsigned char kCopyrightExtensionID				= 0x04;
const unsigned char kSequenceScalableExtensionID		= 0x05;
const unsigned char kPictureDisplayExtensionID			= 0x07;
const unsigned char kPictureCodingExtensionID			= 0x08;
const unsigned char kPictureSpatialScalableExtensionID	= 0x09;
const unsigned char kPictureTemporalScalableExtensionID = 0x0A;

cMPEGVideoFrame::cMPEGVideoFrame(unsigned char * data, unsigned int length, int frameType)
:	cBitStream(data, length),
	frameType(frameType)
{
	memset(&sh, 0, sizeof(tSequenceHeader));
	memset(&se, 0, sizeof(tSequenceExtension));
	memset(&sde, 0, sizeof(tSequenceDisplayExtension));
	memset(&sse, 0, sizeof(tSequenceScalableExtension));
	memset(&goph, 0, sizeof(tGroupOfPicturesHeader));
	memset(&ph, 0, sizeof(tPictureHeader));
	memset(&pce, 0, sizeof(tPictureCodingExtension));
	memset(&qme, 0, sizeof(tQuantMatrixExtension));
	memset(&pde, 0, sizeof(tPictureDisplayExtension));
	memset(&ptse, 0, sizeof(tPictureTemporalScalableExtension));
	memset(&psse, 0, sizeof(tPictureSpatialScalableExtension));
	memset(&ce, 0, sizeof(tCopyrightExtension));
	Parse();
}

cMPEGVideoFrame::~cMPEGVideoFrame()
{
}

cRawVideoFrame * cMPEGVideoFrame::Decode()
{
	AVCodec *codec;
	AVCodecContext *c = NULL;
	int got_picture, len;
	AVFrame *picture;
	cRawVideoFrame * frame = NULL;

	/* find the mpeg1 video decoder */
	codec = avcodec_find_decoder(CODEC_ID_MPEG1VIDEO);
	if (!codec)
	{
		esyslog("codec not found\n");
		return NULL;
	}

	c = avcodec_alloc_context();
	picture = avcodec_alloc_frame();

	/* open it */
	if (avcodec_open(c, codec) < 0)
	{
		esyslog("could not open codec\n");
		return NULL;
	}
    
	len = avcodec_decode_video(c, picture, &got_picture,
							   this->data, this->length);
	if (len < 0)
	{
		esyslog("Error while decoding frame\n");
		return NULL;
	}
	if (got_picture)
	{
		/* the picture is allocated by the decoder. no need to
		   free it */
		frame = new cRawVideoFrame(picture, c->width, c->height);
	}
	else
	{
		len = avcodec_decode_video(c, picture, &got_picture,
								   NULL, 0);
		if (got_picture)
		{
			// the picture is allocated by the decoder. no need to
			// free it
			frame = new cRawVideoFrame(picture, c->width, c->height);
		}
	}

	avcodec_close(c);
	free(c);
	return frame;
}

void cMPEGVideoFrame::Parse()
{
	unsigned long byte;
	while (true)
	{
		if (!NextStartCode())
			break;
		GetBits(24);
		switch(byte = GetBits(8))
		{
		case kPictureStartCode:
			ParsePictureHeader();
			break;
		case kUserDataStartCode:
			break;
		case kSequenceHeaderCode:
			ParseSequenceHeader();
			break;
		case kSequenceErrorCode:
			break;
		case kExtensionStartCode:
			ParseExtension();
			break;
		case kSequenceEndCode:
			break;
		case kGOPStartCode:
			ParseGroupOfPicturesHeader();
			break;
		default:
			if ((byte >= 0x01) && (byte <= 0xAF))
				return;
			break;
		}
	}
}

void cMPEGVideoFrame::ParseExtension()
{
	switch(GetBits(4))
	{
	case kSequenceExtensionID:
		ParseSequenceExtension();
		break;
	case kSequenceDisplayExtensionID:
		ParseSequenceDisplayExtension();
		break;
	case kQuantMatrixExtensionID:
		ParseQuantMatrixExtension();
		break;
	case kSequenceScalableExtensionID:
		ParseSequenceScalableExtension();
		break;
	case kPictureDisplayExtensionID:
		ParsePictureDisplayExtension();
		break;
	case kPictureCodingExtensionID:
		ParsePictureCodingExtension();
		break;
	case kPictureSpatialScalableExtensionID:
		ParsePictureSpatialScalableExtension();
		break;
	case kPictureTemporalScalableExtensionID:
		ParsePictureTemporalScalableExtension();
		break;
	default:
		break;
	}
}

void cMPEGVideoFrame::ParseSequenceHeader()
{
	sh.found = true;

	sh.horizontalSizeValue = GetBits(12);
	sh.verticalSizeValue = GetBits(12);
	sh.aspectRatioInformation = GetBits(4);
	sh.frameRateCode = GetBits(4);
	sh.bitRateValue = GetBits(18) * 400;
	GetBits(1);
	sh.vbvBufferSizeValue = GetBits(10);
	sh.constrainedParametersFlag = GetBits(1);
	// Load Intra Quantiser Matrix
	if ((sh.loadIntraQuantiserMatrix = GetBits(1)))
	{
		for (int i = 0; i < 64; i++)
			sh.intraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
	// Load Non Intra Quantiser Matrix
	if ((sh.loadNonIntraQuantiserMatrix = GetBits(1)))
	{
		for (int i = 0; i < 64; i++)
			sh.nonIntraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
}

void cMPEGVideoFrame::ParseSequenceExtension()
{
	se.found = true;

	se.profileAndLevelIndication = GetBits(8);
	se.progressiveSequence = GetBits(1);
	se.chromaFormat = GetBits(2);
	se.horizontalSizeExtension = GetBits(2);
	se.verticalSizeExtension = GetBits(2);
	se.bitRateExtension = GetBits(12);
	GetBits(1);
	se.vbvBufferSizeExtension = GetBits(8);
	se.lowDelay = GetBits(1);
	se.frameRateExtensionN = GetBits(2);
	se.frameRateExtensionD = GetBits(5);
}

void cMPEGVideoFrame::ParseSequenceDisplayExtension()
{
	sde.found = true;

	sde.videoFormat = GetBits(3);
	sde.colourDescription = GetBits(1);
	if (sde.colourDescription)
	{
		sde.colourPrimaries = GetBits(8);
		sde.transferCharacteristics = GetBits(8);
		sde.matrixCoefficients = GetBits(8);
	}
	sde.displayHorizontalSize = GetBits(14);
	GetBits(1);
	sde.displayVerticalSize = GetBits(14);
}

void cMPEGVideoFrame::ParseSequenceScalableExtension()
{
	sse.found = true;

	sse.scalableMode = GetBits(2);
	sse.layerId = GetBits(4);
	if (sse.scalableMode == 1)
	{
		sse.lowerLayerPredictionHorizontalSize = GetBits(14);
		GetBits(1);
		sse.lowerLayerPredictionVerticalSize = GetBits(14);
		sse.horizontalSubsamplingFactorM = GetBits(5);
		sse.horizontalSubsamplingFactorN = GetBits(5);
		sse.verticalSubsamplingFactorM = GetBits(5);
		sse.verticalSubsamplingFactorN = GetBits(5);
	}
	if (sse.scalableMode == 3)
	{
		sse.pictureMuxEnable = GetBits(1);
		if (sse.pictureMuxEnable)
		{
			sse.muxToProgressiveSequence = GetBits(1);
		}
		sse.pictureMuxOrder = GetBits(3);
		sse.pictureMuxFactor = GetBits(3);
	}
}

void cMPEGVideoFrame::ParseGroupOfPicturesHeader()
{
	goph.found = true;

	goph.dropFrameFlag = GetBits(1);
	goph.timeCodeHours = GetBits(5);
	goph.timeCodeMinutes = GetBits(6);
	GetBits(1);
	goph.timeCodeSeconds = GetBits(6);
	goph.timeCodePictures = GetBits(6);
	goph.closedGop = GetBits(1);
	goph.brokenLink = GetBits(1);
}

void cMPEGVideoFrame::ParsePictureHeader()
{
	ph.found = true;

	ph.temporalReference = GetBits(10);
	ph.pictureCodingType = GetBits(3);
	ph.vbvDelay = GetBits(16);
	if (ph.pictureCodingType == kFrameTypeP ||
		ph.pictureCodingType == kFrameTypeB)
	{
		ph.fullPelForwardVector = GetBits(1);
		ph.forwardFCode = GetBits(3);
	}
	if (ph.pictureCodingType == kFrameTypeB)
	{
		ph.fullPelBackwardVector = GetBits(1);
		ph.backwardFCode = GetBits(3);
	}
	while ((ph.extraBitPicture = GetBits(1)))
	{
		GetBits(8);
	}
}

void cMPEGVideoFrame::ParsePictureCodingExtension()
{
	pce.found = true;

	pce.fCode00 = GetBits(4);
	pce.fCode01 = GetBits(4);
	pce.fCode10 = GetBits(4);
	pce.fCode11 = GetBits(4);
	pce.intraDcPrecision = GetBits(2);
	pce.pictureStructure = GetBits(2);
	pce.topFieldFirst = GetBits(1);
	pce.framePredFrameDct = GetBits(1);
	pce.concealmentMotionVectors = GetBits(1);
	pce.qScaleType = GetBits(1);
	pce.intraVlcFormat = GetBits(1);
	pce.alternateScan = GetBits(1);
	pce.repeatFirstField = GetBits(1);
	pce.chroma420Type = GetBits(1);
	pce.progressiveFrame = GetBits(1);
	pce.compositeDisplayFlag = GetBits(1);
	if (pce.compositeDisplayFlag)
	{
		pce.vAxis = GetBits(1);
		pce.fieldSequence = GetBits(3);
		pce.subCarrier = GetBits(1);
		pce.burstAmplitude = GetBits(7);
		pce.subCarrierPhase = GetBits(8);
	}
}

void cMPEGVideoFrame::ParseQuantMatrixExtension()
{
	int i;

	qme.found = true;

	qme.loadIntraQuantiserMatrix = GetBits(1);
	if (qme.loadIntraQuantiserMatrix)
	{
		for (i = 0; i < 64; i++)
			qme.intraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
	qme.loadNonIntraQuantiserMatrix = GetBits(1);
	if (qme.loadNonIntraQuantiserMatrix)
	{
		for (i = 0; i < 64; i++)
			qme.nonIntraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
	qme.loadChromaIntraQuantiserMatrix = GetBits(1);
	if (qme.loadChromaIntraQuantiserMatrix)
	{
		for (i = 0; i < 64; i++)
			qme.chromaIntraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
	qme.loadChromaNonIntraQuantiserMatrix = GetBits(1);
	if (qme.loadChromaNonIntraQuantiserMatrix)
	{
		for (i = 0; i < 64; i++)
			qme.chromaNonIntraQuantiserMatrix[i] = (unsigned char) GetBits(8);
	}
}

void cMPEGVideoFrame::ParsePictureDisplayExtension()
{
	pde.found = true;

	if ((se.progressiveSequence == 1) ||
		((pce.pictureStructure == 1) || (pce.pictureStructure == 2)))
	{
		pde.numberOfFrameCentreOffsets = 1;
	}
	else
	{
		if (pce.repeatFirstField == 1)
		{
			pde.numberOfFrameCentreOffsets = 3;
		}
		else
		{
			pde.numberOfFrameCentreOffsets = 2;
		}
	}
	for (int i = 0; i < (int) pde.numberOfFrameCentreOffsets; i++)
	{
		pde.frameCentreHorizontalOffset[i] = GetBits(16);
		GetBits(1);
		pde.frameCentreVerticalOffset[i] = GetBits(16);
		GetBits(1);
	}
}

void cMPEGVideoFrame::ParsePictureTemporalScalableExtension()
{
	ptse.found = true;

	ptse.referenceSelectCode = GetBits(2);
	ptse.forwardTemporalReference = GetBits(10);
	GetBits(1);
	ptse.backwardTemporalReference = GetBits(10);
}

void cMPEGVideoFrame::ParsePictureSpatialScalableExtension()
{
	psse.found = true;

	psse.lowerLayerTemporalReference = GetBits(10);
	GetBits(1);
	psse.lowerLayerHorizontalOffset = GetBits(15);
	GetBits(1);
	psse.lowerLayerVerticalOffset = GetBits(15);
	psse.spatialTemporalWeightCodeTableIndex = GetBits(2);
	psse.lowerLayerProgressiveFrame = GetBits(1);
	psse.lowerLayerDeinterlacedFieldSelect = GetBits(1);
}

void cMPEGVideoFrame::ParseCopyrightExtension()
{
	ce.found = true;

	ce.copyrightFlag = GetBits(1);
	ce.copyrightIdentifier = GetBits(8);
	ce.originalOrCopy = GetBits(1);
	GetBits(7);
	GetBits(1);
	ce.copyrightNumber1 = GetBits(20);
	GetBits(1);
	ce.copyrightNumber2 = GetBits(22);
	GetBits(1);
	ce.copyrightNumber3 = GetBits(22);
}

bool cMPEGVideoFrame::NextStartCode()
{
	AlignBits();
	while (NextBits(24) != 0x000001)
	{
		GetBits(8);
		if (IsOnEnd())
			return false;
	}
	return true;
}
