/*
 * video.h: Video frame handling
 *
 * See the main source file 'main.cpp' for copyright information and
 * how to reach the author.
 */

#ifndef _VIDEO_H_
#define _VIDEO_H_

#include "stream.h"
#include <stdio.h>
extern "C" {
#include <ffmpeg/avcodec.h>
}

extern FILE *rawYVideoFile;
extern FILE *rawUVideoFile;
extern FILE *rawVVideoFile;
extern FILE *mpegVideoFile;

/**
 * A class representing an uncompressed video frame. The data is stored in YUV format.
 */
class cRawVideoFrame
{
public:
	/**
	 * Constructs the video frame. Picture data is copied.
	 *
	 * @param picture pointer to picture structure filled by decoder
	 * @param width width of picture in pixels
	 * @param height height of picture in pixels
	 */
	cRawVideoFrame(AVFrame *picture, int width, int height);
	/**
	 * Destructs the video frame.
	 */
	~cRawVideoFrame();
	/**
	 * Retrieves the data of component i.
	 *
	 * @param i number of component (0 = Y, 1 = U, 2 = V)
	 *
	 * @return a pointer to the data of component i
	 */
	unsigned char * PictureData(int i);
	/**
	 * Returns the length of a line of component i in memory.
	 *
	 * @param i number of component (0 = Y, 1 = U, 2 = V)
	 * @return length in bytes
	 */
	int LineSize(int i);
	/**
	 * Returns the width of the video frame.
	 *
	 * @return width of video frame in pixels
	 */
	int Width() { return width; }
	/**
	 * Returns the height of the video frame.
	 *
	 * @return height of video frame in pixels
	 */
	int Height() { return height; }
private:
	AVPicture picture;
	int width;
	int height;
};

typedef struct
{
	bool found;
	unsigned int horizontalSizeValue;
	unsigned int verticalSizeValue;
	unsigned int aspectRatioInformation;
	unsigned int frameRateCode;
	unsigned int bitRateValue;
	unsigned int vbvBufferSizeValue;
	unsigned int constrainedParametersFlag;
	unsigned int loadIntraQuantiserMatrix;
	unsigned char intraQuantiserMatrix[64];
	unsigned int loadNonIntraQuantiserMatrix;
	unsigned char nonIntraQuantiserMatrix[64];
} tSequenceHeader;

typedef struct
{
	bool found;
	unsigned int profileAndLevelIndication;
	unsigned int progressiveSequence;
	unsigned int chromaFormat;
	unsigned int horizontalSizeExtension;
	unsigned int verticalSizeExtension;
	unsigned int bitRateExtension;
	unsigned int vbvBufferSizeExtension;
	unsigned int lowDelay;
	unsigned int frameRateExtensionN;
	unsigned int frameRateExtensionD;
} tSequenceExtension;

typedef struct
{
	bool found;
	unsigned int videoFormat;
	unsigned int colourDescription;
	unsigned int colourPrimaries;
	unsigned int transferCharacteristics;
	unsigned int matrixCoefficients;
	unsigned int displayHorizontalSize;
	unsigned int displayVerticalSize;
} tSequenceDisplayExtension;

typedef struct
{
	bool found;
	unsigned int scalableMode;
	unsigned int layerId;
	unsigned int lowerLayerPredictionHorizontalSize;
	unsigned int lowerLayerPredictionVerticalSize;
	unsigned int horizontalSubsamplingFactorM;
	unsigned int horizontalSubsamplingFactorN;
	unsigned int verticalSubsamplingFactorM;
	unsigned int verticalSubsamplingFactorN;
	unsigned int pictureMuxEnable;
	unsigned int muxToProgressiveSequence;
	unsigned int pictureMuxOrder;
	unsigned int pictureMuxFactor;
} tSequenceScalableExtension;

typedef struct
{
	bool found;
	unsigned int dropFrameFlag;
	unsigned int timeCodeHours;
	unsigned int timeCodeMinutes;
	unsigned int timeCodeSeconds;
	unsigned int timeCodePictures;
	unsigned int closedGop;
	unsigned int brokenLink;
} tGroupOfPicturesHeader;

typedef struct
{
	bool found;
	unsigned int temporalReference;
	unsigned int pictureCodingType;
	unsigned int vbvDelay;
	unsigned int fullPelForwardVector;
	unsigned int forwardFCode;
	unsigned int fullPelBackwardVector;
	unsigned int backwardFCode;
	unsigned int extraBitPicture;
} tPictureHeader;

typedef struct
{
	bool found;
	unsigned int fCode00;
	unsigned int fCode01;
	unsigned int fCode10;
	unsigned int fCode11;
	unsigned int intraDcPrecision;
	unsigned int pictureStructure;
	unsigned int topFieldFirst;
	unsigned int framePredFrameDct;
	unsigned int concealmentMotionVectors;
	unsigned int qScaleType;
	unsigned int intraVlcFormat;
	unsigned int alternateScan;
	unsigned int repeatFirstField;
	unsigned int chroma420Type;
	unsigned int progressiveFrame;
	unsigned int compositeDisplayFlag;
	unsigned int vAxis;
	unsigned int fieldSequence;
	unsigned int subCarrier;
	unsigned int burstAmplitude;
	unsigned int subCarrierPhase;
} tPictureCodingExtension;

typedef struct
{
	bool found;
	unsigned int loadIntraQuantiserMatrix;
	unsigned char intraQuantiserMatrix[64];
	unsigned int loadNonIntraQuantiserMatrix;
	unsigned char nonIntraQuantiserMatrix[64];
	unsigned int loadChromaIntraQuantiserMatrix;
	unsigned char chromaIntraQuantiserMatrix[64];
	unsigned int loadChromaNonIntraQuantiserMatrix;
	unsigned char chromaNonIntraQuantiserMatrix[64];
} tQuantMatrixExtension;

typedef struct
{
	bool found;
	unsigned int numberOfFrameCentreOffsets;
	unsigned int frameCentreHorizontalOffset[3];
	unsigned int frameCentreVerticalOffset[3];
} tPictureDisplayExtension;

typedef struct
{
	bool found;
	unsigned int referenceSelectCode;
	unsigned int forwardTemporalReference;
	unsigned int backwardTemporalReference;
} tPictureTemporalScalableExtension;

typedef struct
{
	bool found;
	unsigned int lowerLayerTemporalReference;
	unsigned int lowerLayerHorizontalOffset;
	unsigned int lowerLayerVerticalOffset;
	unsigned int spatialTemporalWeightCodeTableIndex;
	unsigned int lowerLayerProgressiveFrame;
	unsigned int lowerLayerDeinterlacedFieldSelect;
} tPictureSpatialScalableExtension;

typedef struct
{
	bool found;
	unsigned int copyrightFlag;
	unsigned int copyrightIdentifier;
	unsigned int originalOrCopy;
	unsigned int copyrightNumber1;
	unsigned int copyrightNumber2;
	unsigned int copyrightNumber3;
} tCopyrightExtension;

/**
 * A class representing a compressed MPEG video frame.
 */
class cMPEGVideoFrame
: public cBitStream
{
public:    
	/**
	 * Constructs the video frame. Parses the data and fills header structures.
	 *
	 * @param data pointer to frame data
	 * @param length length of frame data
	 * @param frameType type of frame (I, P, B, D)
	 */
	cMPEGVideoFrame(unsigned char * data, unsigned int length, int frameType);
	/**
	 * Destructs the video frame. Actually does nothing.
	 */
	~cMPEGVideoFrame();
	/**
	 * Decodes the video frame using libavcodec. This only works for I-Frames.
	 *
	 * @return a pointer to the decompressed video frame
	 */
	cRawVideoFrame * Decode();
	/**
	 * Retrieves the type of the video frame.
	 *
	 * @return frame type (1 = I-Frame, 2 = P-Frame, 3 = B-Frame, 4 = D-Frame)
	 */
	int FrameType() { return frameType; }
	/**
	 * Retrieves the sequence header data of the video frame.
	 *
	 * @return structure of type tSequenceHeader containing the data
	 */
	tSequenceHeader * SequenceHeader() { return &sh; }
	/**
	 * Retrieves the sequence extension data of the video frame.
	 *
	 * @return structure of type tSequenceExtension containing the data
	 */
	tSequenceExtension * SequenceExtension() { return &se; }
	/**
	 * Retrieves the sequence display extension data of the video frame.
	 *
	 * @return structure of type tSequenceDisplayExtension containing the data
	 */
	tSequenceDisplayExtension * SequenceDisplayExtension() { return &sde; }
	/**
	 * Retrieves the sequence scalable extension data of the video frame.
	 *
	 * @return structure of type tSequenceScalableExtension containing the data
	 */
	tSequenceScalableExtension * SequenceScalableExtension() { return &sse; }
	/**
	 * Retrieves the group of pictures header data of the video frame.
	 *
	 * @return structure of type tGroupOfPicturesHeader containing the data
	 */
	tGroupOfPicturesHeader * GroupOfPicturesHeader() { return &goph; }
	/**
	 * Retrieves the picture header data of the video frame.
	 *
	 * @return structure of type tPictureHeader containing the data
	 */
	tPictureHeader * PictureHeader() { return &ph; }
	/**
	 * Retrieves the picture coding extension data of the video frame.
	 *
	 * @return structure of type tPictureCodingExtension containing the data
	 */
	tPictureCodingExtension * PictureCodingExtension() { return &pce; }
	/**
	 * Retrieves the quant matrix extension data of the video frame.
	 *
	 * @return structure of type tQuantMatrixExtension containing the data
	 */
	tQuantMatrixExtension * QuantMatrixExtension() { return &qme; }
	/**
	 * Retrieves the picture display extension data of the video frame.
	 *
	 * @return structure of type tPictureDisplayExtension containing the data
	 */
	tPictureDisplayExtension * PictureDisplayExtension() { return &pde; }
	/**
	 * Retrieves the picture temporal scalable extension data of the video frame.
	 *
	 * @return structure of type tPictureTemporalScalableExtension containing the data
	 */
	tPictureTemporalScalableExtension * PictureTemporalScalableExtension() { return &ptse; }
	/**
	 * Retrieves the picture spatial scalable extension data of the video frame.
	 *
	 * @return structure of type tPictureSpatialScalableExtension containing the data
	 */
	tPictureSpatialScalableExtension * PictureSpatialScalableExtension() { return &psse; }
	/**
	 * Retrieves the copyright extension data of the video frame.
	 *
	 * @return structure of type tCopyrightExtension containing the data
	 */
	tCopyrightExtension * CopyrightExtension() { return &ce; }
private:
	void Parse();
	void ParseExtension();
	void ParseSequenceHeader();
	void ParseSequenceExtension();
	void ParseSequenceDisplayExtension();
	void ParseSequenceScalableExtension();
	void ParseGroupOfPicturesHeader();
	void ParsePictureHeader();
	void ParsePictureCodingExtension();
	void ParseQuantMatrixExtension();
	void ParsePictureDisplayExtension();
	void ParsePictureTemporalScalableExtension();
	void ParsePictureSpatialScalableExtension();
	void ParseCopyrightExtension();
	/**
	 * Searches next start code in stream. Returns true if found and false if not found.
	 *
	 * @return true if start code found and false if no start code found
	 */
	bool NextStartCode();
    int frameType;
	tSequenceHeader sh;
	tSequenceExtension se;
	tSequenceDisplayExtension sde;
	tSequenceScalableExtension sse;
	tGroupOfPicturesHeader goph;
	tPictureHeader ph;
	tPictureCodingExtension pce;
	tQuantMatrixExtension qme;
	tPictureDisplayExtension pde;
	tPictureTemporalScalableExtension ptse;
	tPictureSpatialScalableExtension psse;
	tCopyrightExtension ce;
};

const unsigned int kFrameTypeI = 1;
const unsigned int kFrameTypeP = 2;
const unsigned int kFrameTypeB = 3;
const unsigned int kFrameTypeD = 4;

#endif

